from io import BytesIO
from typing import Mapping, Union
from urllib.parse import urlparse

from werkzeug import Request, Response
from werkzeug.datastructures import Headers
from werkzeug.test import EnvironBuilder

from .client import HttpClient, SimpleRequestsClient
from .request import get_raw_path, restore_payload, set_environment_headers


def forward(
    request: Request,
    forward_base_url: str,
    forward_path: str = None,
    headers: Union[Headers, Mapping[str, str]] = None,
) -> Response:
    """
    Convenience method that creates a new Proxy and immediately calls proxy.forward(...). See ``Proxy`` for more
    information.
    """
    with Proxy(forward_base_url=forward_base_url) as proxy:
        return proxy.forward(request, forward_path=forward_path, headers=headers)


class Proxy(HttpClient):
    preserve_host: bool

    def __init__(
        self, forward_base_url: str, client: HttpClient = None, preserve_host: bool = True
    ):
        """
        Creates a new HTTP Proxy which can be used to forward incoming requests according to the configuration.

        :param forward_base_url: the base url (backend) to forward the requests to.
        :param client: the HTTP Client used to make the requests
        :param preserve_host: True to ensure that the Host header of the incoming request is preserved.
                              If False, then the Host header will be set to the Host from the perspective of the Proxy.
        """
        self.forward_base_url = forward_base_url
        self.client = client or SimpleRequestsClient()
        self.preserve_host = preserve_host

    def request(self, request: Request, server: str | None = None) -> Response:
        """
        Compatibility with HttpClient interface. A call is equivalent to ``Proxy.forward(request, None, None)``.

        :param request: the request to proxy
        :param server: ignored for a proxy, since the server is already set by `forward_base_url`.
        :return: the proxied response
        """
        return self.forward(request)

    def forward(
        self,
        request: Request,
        forward_path: str = None,
        headers: Union[Headers, Mapping[str, str]] = None,
    ) -> Response:
        """
        Uses the client to forward the given request according to the proxy's configuration.

        :param request: the base request to forward (with the original URL and path data)
        :param forward_path: the path to forward the request to. if set, the original path will be replaced completely,
            otherwise the original path will be used
        :param headers: additional custom headers to send as part of the proxy request
        :return: the proxied response
        """
        headers = Headers(headers) if headers else Headers()

        if client_ip := request.remote_addr:
            if xff := request.headers.get("X-Forwarded-For"):
                headers["X-Forwarded-For"] = f"{xff}, {client_ip}"
            else:
                headers["X-Forwarded-For"] = f"{client_ip}"

        if forward_path is None:
            forward_path = get_raw_path(request)
        if forward_path:
            forward_path = "/" + forward_path.lstrip("/")

        proxy_request = _copy_request(request, self.forward_base_url, forward_path, headers)

        if self.preserve_host and "Host" in request.headers:
            proxy_request.headers["Host"] = request.headers["Host"]

        target = urlparse(self.forward_base_url)
        return self.client.request(proxy_request, server=f"{target.scheme}://{target.netloc}")

    def close(self):
        self.client.close()


class ProxyHandler:
    """
    A dispatcher Handler which can be used in a ``Router[Handler]`` that proxies incoming requests according to the
    configuration.

    The Handler is expected to be used together with a route that uses a ``path`` parameter named ``path`` in the URL.
    Fir example: if you want to forward all requests from ``/foobar/<path>`` to ``http://localhost:8080/v1/<path>``,
    you would do the following::

        router = Router(dispatcher=handler_dispatcher())
        router.add("/foobar/<path:path>", ProxyHandler("http://localhost:8080/v1")

    This is similar to the common nginx configuration where proxy_pass is a URI::

        location /foobar {
            proxy_pass http://localhost:8080/v1/;
        }
    """

    def __init__(self, forward_base_url: str, client: HttpClient = None):
        """
        Creates a new Proxy with the given ``forward_base_url`` (see ``Proxy``).

        :param forward_base_url: the base url (backend) to forward the requests to.
        :param client: the HTTP Client used to make the requests
        """
        self.proxy = Proxy(forward_base_url=forward_base_url, client=client)

    def __call__(self, request: Request, **kwargs) -> Response:
        return self.proxy.forward(request, forward_path=kwargs.get("path", ""))

    def close(self):
        self.proxy.close()


def _copy_request(
    request: Request,
    base_url: str = None,
    path: str = None,
    headers: Union[Headers, Mapping[str, str]] = None,
) -> Request:
    """
    Creates a new request from the given one that can be used to perform a proxy call.

    :param request: the original request
    :param base_url: the url to forward the request to (e.g., http://localhost:8080)
    :param path: the path to forward the request to (e.g., /foobar), if set to None, the original path will be used
    :param headers: optional headers to overwrite
    :return: a new request with slightly modified underlying environment but the same data stream
    """
    # ensure that the headers in the env are set on the environment
    # FIXME: we should preserve header casing like we do with the `asgi.headers` property in the  asgi/wsgi bridge to
    #  pass through the raw headers.
    set_environment_headers(request.environ, request.headers)
    builder = EnvironBuilder.from_environ(request.environ)

    if base_url:
        builder.base_url = base_url
        builder.headers["Host"] = builder.host

    if path is not None:
        builder.path = path

    if headers:
        builder.headers.update(headers)

    # FIXME: unfortunately, EnvironBuilder expects the input stream to be seekable, but we don't have that when using
    #  the asgi/wsgi bridge. we need a better way of dealing with IO!
    data = restore_payload(request)
    builder.input_stream = BytesIO(data)
    builder.content_length = len(data)
    # Since the payload is completely restored, the proxy forwarding is not streamed.
    # Therefore, we need to remove a potential "chunked" Transfer-Encoding
    if builder.headers.get("Transfer-Encoding", None) == "chunked":
        builder.headers.pop("Transfer-Encoding")

    new_request = builder.get_request()

    # explicitly set the path in the environment and in the newly created request
    if path is not None:
        new_request.environ["RAW_URI"] = path or "/"

    # copy headers s.t. they are no longer immutable (by default, EnvironHeaders are used)
    new_request.headers = Headers(new_request.headers)

    return new_request



"""
Bindings to serve rolo through Twisted.
"""
import logging
import typing as t
from io import BytesIO
from queue import Empty, Queue

from twisted.internet import reactor
from twisted.internet.protocol import Protocol
from twisted.protocols.policies import ProtocolWrapper
from twisted.python.components import proxyForInterface
from twisted.web.http import HTTPChannel, _GenericHTTPChannelProtocol
from twisted.web.http_headers import Headers as TwistedHeaders
from twisted.web.resource import IResource
from twisted.web.server import NOT_DONE_YET, Request, Site
from twisted.web.server import Request as TwistedRequest
from twisted.web.wsgi import WSGIResource, _WSGIResponse, _wsgiString
from werkzeug.datastructures import Headers
from wsproto import ConnectionType, WSConnection, events
from zope.interface import implementer

from rolo.gateway import Gateway
from rolo.gateway.wsgi import WsgiGateway
from rolo.websocket import (
    WebSocketDisconnectedError,
    WebSocketEnvironment,
    WebSocketListener,
    WebSocketProtocolError,
    WebSocketRequest,
)
from rolo.websocket import (
    adapter as rolows,
)

if t.TYPE_CHECKING:
    from _typeshed.wsgi import WSGIEnvironment


LOG = logging.getLogger(__name__)


def to_flat_header_list(headers: TwistedHeaders) -> list[tuple[bytes, bytes]]:
    result = []
    for k, vs in headers.getAllRawHeaders():
        for v in vs:
            result.append((k, v))
    return result


def update_wsgi_environment(environ: "WSGIEnvironment", request: TwistedRequest):
    """
    Update the pre-populated WSGI environment with additional data, needed by rolo, from the webserver
    request object.

    :param environ: the environment to update
    :param request: the webserver request object
    """
    # store raw headers
    environ["rolo.headers"] = to_flat_header_list(request.requestHeaders)

    # TODO: check if twisted input streams are really properly terminated
    # this is needed for streaming requests
    environ["wsgi.input_terminated"] = True

    # create RAW_URI and REQUEST_URI
    environ["REQUEST_URI"] = request.uri.decode("utf-8")
    environ["RAW_URI"] = request.uri.decode("utf-8")
    # client addr/port
    addr = request.getClientAddress()
    environ["REMOTE_ADDR"] = addr.host
    environ["REMOTE_PORT"] = str(addr.port)


def to_websocket_environment(request: Request) -> WebSocketEnvironment:
    """
    Creates a pseudo WSGI environment to be used for the rolo WebsocketRequest. Partially copied from twisted.web.wsgi.

    :param request: the twisted webserver request
    :return: a WSGI-like environment for rolo
    """
    if request.prepath:
        scriptName = b"/" + b"/".join(request.prepath)
    else:
        scriptName = b""

    if request.postpath:
        pathInfo = b"/" + b"/".join(request.postpath)
    else:
        pathInfo = b""

    parts = request.uri.split(b"?", 1)
    if len(parts) == 1:
        queryString = b""
    else:
        queryString = parts[1]

    environ = {
        "REQUEST_METHOD": "WEBSOCKET",
        "REMOTE_ADDR": _wsgiString(request.getClientAddress().host),
        "REMOTE_PORT": _wsgiString(str(request.getClientAddress().port)),
        "SCRIPT_NAME": _wsgiString(scriptName),
        "PATH_INFO": _wsgiString(pathInfo),
        "QUERY_STRING": _wsgiString(queryString),
        "CONTENT_TYPE": _wsgiString(request.getHeader(b"content-type") or ""),
        "CONTENT_LENGTH": _wsgiString(request.getHeader(b"content-length") or ""),
        "SERVER_NAME": _wsgiString(request.getRequestHostname()),
        "SERVER_PORT": _wsgiString(str(request.getHost().port)),
        "SERVER_PROTOCOL": _wsgiString(request.clientproto),
        "REQUEST_URI": request.uri.decode("utf-8"),
        "RAW_URI": request.uri.decode("utf-8"),
    }

    # store raw headers for rolo
    environ["rolo.headers"] = to_flat_header_list(request.requestHeaders)

    # WSGI headers
    for name, values in request.requestHeaders.getAllRawHeaders():
        name = "HTTP_" + _wsgiString(name).upper().replace("-", "_")
        environ[name] = ",".join(_wsgiString(v) for v in values).replace("\n", " ")

    environ.update(
        {
            "wsgi.version": (1, 0),
            "wsgi.url_scheme": request.isSecure() and "https" or "http",
            "wsgi.run_once": False,
            "wsgi.multithread": True,
            "wsgi.multiprocess": False,
            "wsgi.errors": BytesIO(),
            "wsgi.input": BytesIO(),
        }
    )

    return environ


class TwistedRequestAdapter(TwistedRequest):
    """
    Custom twisted server Request object to handle header casing.
    """

    rawHeaderList: list[tuple[bytes, bytes]]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # instantiate case mappings, these are used by `getAllRawHeaders` to restore casing
        # by default, they are class attributes, so we would override them globally
        self.requestHeaders._caseMappings = dict(self.requestHeaders._caseMappings)
        self.responseHeaders._caseMappings = dict(self.responseHeaders._caseMappings)


class HeaderPreservingHTTPChannel(HTTPChannel):
    """
    Special HTTPChannel implementation that uses ``Headers._caseMappings`` to retain header casing both for
    request headers (server -> WSGI), and  response headers (WSGI -> client).
    """

    requestFactory = TwistedRequestAdapter

    @staticmethod
    def protocol_factory():
        return _GenericHTTPChannelProtocol(HeaderPreservingHTTPChannel())

    def headerReceived(self, line):
        if not super().headerReceived(line):
            return False

        # remember casing of headers for requests
        header, data = line.split(b":", 1)
        request: TwistedRequestAdapter = self.requests[-1]
        request.requestHeaders._caseMappings[header.lower()] = header
        return True

    def writeHeaders(self, version, code, reason, headers):
        """Alternative implementation that writes the raw headers instead of sanitized versions."""
        responseLine = version + b" " + code + b" " + reason + b"\r\n"
        headerSequence = [responseLine]

        for name, value in headers:
            line = name + b": " + value + b"\r\n"
            headerSequence.append(line)

        headerSequence.append(b"\r\n")
        self.transport.writeSequence(headerSequence)

    def isSecure(self):
        # used to determine the WSGI url scheme (http vs https)
        try:
            # ``self.transport`` will be a ``TLSMultiplexer`` instance in our case
            return self.transport.isSecure()
        except AttributeError:
            return super().isSecure()


class HeaderPreservingWSGIResponse(_WSGIResponse):
    def __init__(self, reactor, threadpool, application, request):
        super().__init__(reactor, threadpool, application, request)
        update_wsgi_environment(self.environ, request)

    def startResponse(self, *args, **kwargs):
        result = super().startResponse(*args, **kwargs)
        # before starting the WSGI response, make sure we store the raw case mappings into the response
        # headers
        for header, _ in self.headers:
            header = header.encode("latin-1")
            self.request.responseHeaders._caseMappings[header.lower()] = header
        return result


class HeaderPreservingWSGIResource(WSGIResource):
    def render(self, request):
        response = HeaderPreservingWSGIResponse(
            self._reactor, self._threadpool, self._application, request
        )
        response.start()
        return NOT_DONE_YET


@implementer(IResource)
class WebsocketResourceDecorator(proxyForInterface(IResource)):
    """
    Wrapper around a ``WSGIResource`` that intercepts websocket requests, and calls the
    ``WebSocketListener`` in a separate thread, similar to how ``WSGIResource`` dispatches its requests.
    """

    original: WSGIResource
    isLeaf = True

    def __init__(
        self,
        original: WSGIResource,
        websocketListener: WebSocketListener,
    ):
        super().__init__(original)
        self.websocketListener = websocketListener

    def render(self, request: Request):
        if upgrade := request.getHeader("upgrade"):
            if upgrade.lower() == "websocket":
                self._processWebsocket(request)
                return NOT_DONE_YET

        return super().render(request)

    def _processWebsocket(self, request: Request):
        channel = WebSocketChannel(request)
        if isinstance(request.channel.transport, ProtocolWrapper):
            request.transport.wrappedProtocol = channel
        else:
            request.transport.protocol = channel

        channel.initiateUpgrade()

        environment = to_websocket_environment(request)
        environment["rolo.websocket"] = TwistedWebSocketAdapter(channel)
        # WSGIResource also dispatches requests through the threadpool
        self.original._threadpool.callInThread(self.websocketListener, environment)


class WebSocketChannel(Protocol):
    """
    Websocket protocol implementation over twisted. Note this is a ``twisted.internet.Protocol``, not a
    Python Protocol class.
    """

    eventQueue: Queue[events.Event]

    def __init__(self, request: Request):
        self.request = request
        self.wsproto = WSConnection(ConnectionType.SERVER)
        self.eventQueue = Queue()

    @property
    def closed(self):
        return self.request.finished

    def initiateUpgrade(self):
        headers = [(k, v) for k, vs in self.request.requestHeaders.getAllRawHeaders() for v in vs]
        self.wsproto.initiate_upgrade_connection(headers, self.request.path)

        for event in self.wsproto.events():
            self.eventQueue.put(event)
            if isinstance(event, events.CloseConnection):
                self.close()

    def connectionLost(self, reason):
        self.close()

    def dataReceived(self, data: bytes) -> None:
        self.wsproto.receive_data(data)
        for event in self.wsproto.events():
            if isinstance(event, events.Ping):
                self.wsSend(events.Pong(event.payload))
                continue
            # TODO: filter other event types that are not expected by WebSocketAdapter
            if isinstance(event, events.CloseConnection):
                self.close()
            self.eventQueue.put_nowait(event)

    def wsSend(self, event: events.Event):
        request = self.request
        if request.finished:
            return
        data = self.wsproto.send(event)
        request.transport.write(data)

    def wsReject(
        self,
        statusCode: int,
        extraHeaders: Headers,
        body: t.Iterator[bytes] | None = None,
    ):
        # this sends an HTTP response back to the client before the request was upgraded. we could also use
        # ``self.wsSend(events.RejectConnection(statusCode, ...))`` which would write the HTTP response
        # generated by wsproto to the transport, but instead we re-use twisted's request/response mechanism
        # which is cleaner here, though perhaps inconsistent with the rest of the implementation.
        # TODO: set default twisted headers
        request = self.request

        request.setResponseCode(statusCode)
        for k, v in extraHeaders.to_wsgi_list():
            request.responseHeaders.addRawHeader(k, v)

        if body:
            for b in body:
                request.write(b)

        self.close()

    def wsClose(self, code: int = 1000, reason: t.Optional[str] = None):
        try:
            self.wsSend(events.CloseConnection(code, reason))
        finally:
            self.close()

    def close(self):
        if not self.request.finished:
            self.request.finish()
            # special internal poison pill
            self.eventQueue.put_nowait(events.CloseConnection(None))


class TwistedWebSocketAdapter(rolows.WebSocketAdapter):
    channel: WebSocketChannel

    def __init__(self, channel: WebSocketChannel):
        self.channel = channel

    def receive(self, timeout: float = None) -> rolows.CreateConnection | rolows.Message:
        try:
            event = self.channel.eventQueue.get(timeout=timeout)
        except Empty as e:
            raise TimeoutError(
                f"Timeout error while reading events from websocket after {timeout}s"
            ) from e

        if isinstance(event, events.Request):
            return rolows.CreateConnection()
        if isinstance(event, events.BytesMessage):
            return rolows.BytesMessage(event.data)
        elif isinstance(event, events.TextMessage):
            return rolows.TextMessage(event.data)
        elif isinstance(event, events.CloseConnection):
            raise WebSocketDisconnectedError(event.code)
        else:
            raise WebSocketProtocolError(f"Unexpected event type {event.__class__.__name__}")

    def send(self, event: rolows.Message, timeout: float = None):
        if isinstance(event, rolows.TextMessage):
            self.channel.wsSend(events.TextMessage(event.data))
        elif isinstance(event, rolows.BytesMessage):
            self.channel.wsSend(events.BytesMessage(event.data))
        else:
            raise TypeError(f"Unexpected event type {event.__class__.__name__}")

    def reject(
        self,
        status_code: int,
        headers: Headers = None,
        body: t.Iterable[bytes] = None,
        timeout: float = None,
    ):
        self.channel.wsReject(status_code, headers, body)

    def accept(
        self,
        subprotocol: str = None,
        extensions: list[str] = None,
        extra_headers: Headers = None,
        timeout: float = None,
    ):
        if extra_headers:
            raw_headers = [
                (k.encode("latin-1"), v.encode("latin-1")) for k, v in extra_headers.to_wsgi_list()
            ]
        else:
            raw_headers = []

        # TODO: extensions
        event = events.AcceptConnection(subprotocol, extensions=[], extra_headers=raw_headers)
        self.channel.wsSend(event)

    def close(self, code: int = 1001, reason: str = None, timeout: float = None):
        if not self.channel.closed:
            self.channel.wsClose(code, reason)


class GatewayResource(proxyForInterface(IResource)):
    """
    Compound ``Resource`` implementation to serve a Gateway through a ``Site``.
    """

    def __init__(self, gateway: Gateway, reactor, threadpool):
        self.gateway = gateway
        super().__init__(
            WebsocketResourceDecorator(
                original=HeaderPreservingWSGIResource(reactor, threadpool, WsgiGateway(gateway)),
                websocketListener=WebSocketRequest.listener(gateway.accept),
            )
        )


class TwistedGateway(Site):
    """
    Expose a Gateway as a ``twisted.web.server.Site`` which is a ProtocolFactory that can be served via
    ``reactor.listenTCP``.
    """

    def __init__(self, gateway: Gateway):
        super().__init__(
            GatewayResource(gateway, reactor, reactor.getThreadPool()), TwistedRequestAdapter
        )
        self.protocol = HeaderPreservingHTTPChannel.protocol_factory



import json
import logging
import threading
import time
from concurrent.futures import ThreadPoolExecutor
from queue import Queue
from threading import Thread
from typing import List

import pytest
import requests
import websocket
from werkzeug import Request, Response

from rolo.asgi import ASGILifespanListener, ASGIWebSocketAdapter, WebSocketEnvironment

LOG = logging.getLogger(__name__)


def test_serve_asgi_adapter(serve_asgi_adapter):
    request_list: List[Request] = []

    @Request.application
    def app(request: Request) -> Response:
        request_list.append(request)
        return Response("ok", 200)

    server = serve_asgi_adapter(app)

    response0 = requests.get(server.url + "/foobar?foo=bar", headers={"x-amz-target": "testing"})
    assert response0.ok
    assert response0.text == "ok"

    response1 = requests.get(server.url + "/compute", data='{"foo": "bar"}')
    assert response1.ok
    assert response1.text == "ok"

    request0 = request_list[0]
    assert request0.path == "/foobar"
    assert request0.query_string == b"foo=bar"
    assert request0.full_path == "/foobar?foo=bar"
    assert request0.headers["x-amz-target"] == "testing"
    assert dict(request0.args) == {"foo": "bar"}

    request1 = request_list[1]
    assert request1.path == "/compute"
    assert request1.get_data() == b'{"foo": "bar"}'


def test_requests_are_not_blocking_the_server(serve_asgi_adapter):
    queue = Queue()

    @Request.application
    def app(request: Request) -> Response:
        time.sleep(1)
        queue.put_nowait(request)
        return Response("ok", 200)

    server = serve_asgi_adapter(app)

    then = time.time()

    Thread(target=requests.get, args=(server.url,)).start()
    Thread(target=requests.get, args=(server.url,)).start()
    Thread(target=requests.get, args=(server.url,)).start()
    Thread(target=requests.get, args=(server.url,)).start()

    # get the four responses
    queue.get(timeout=5)
    queue.get(timeout=5)
    queue.get(timeout=5)
    queue.get(timeout=5)

    assert (time.time() - then) < 4, "requests did not seem to be parallelized"


def test_chunked_transfer_encoding_response(serve_asgi_adapter):
    # this test makes sure that creating a response with a generator automatically creates a
    # transfer-encoding=chunked response

    @Request.application
    def app(_request: Request) -> Response:
        def _gen():
            yield "foo"
            yield "bar\n"
            yield "baz\n"

        return Response(_gen(), 200)

    server = serve_asgi_adapter(app)

    response = requests.get(server.url)

    assert response.headers["Transfer-Encoding"] == "chunked"

    it = response.iter_lines()

    assert next(it) == b"foobar"
    assert next(it) == b"baz"


def test_chunked_transfer_encoding_client_timeout(serve_asgi_adapter):
    # this test makes sure that creating a response with a generator automatically creates a
    # transfer-encoding=chunked response

    generator_exited = threading.Event()
    continue_request = threading.Event()

    @Request.application
    def app(_request: Request) -> Response:
        def _gen():
            try:
                yield "foo"
                yield "bar\n"
                continue_request.wait()
                # only three are needed, let's send some more to make sure
                for _ in range(10):
                    yield "baz\n"
            except GeneratorExit:
                generator_exited.set()

        return Response(_gen(), 200)

    server = serve_asgi_adapter(app)

    with requests.get(server.url, stream=True) as response:
        assert response.headers["Transfer-Encoding"] == "chunked"

        it = response.iter_lines()

        assert next(it) == b"foobar"

    # request is now closed, continue the response generator
    continue_request.set()
    # this flag is only set when generator is exited
    assert generator_exited.wait(timeout=10)


def test_chunked_transfer_encoding_request(serve_asgi_adapter):
    request_list: List[Request] = []

    @Request.application
    def app(request: Request) -> Response:
        request_list.append(request)

        stream = request.stream
        data = bytearray()

        for i, item in enumerate(stream):
            data.extend(item)

            if i == 0:
                assert item == b"foobar\n"
            if i == 1:
                assert item == b"baz"

        return Response(data.decode("utf-8"), 200)

    server = serve_asgi_adapter(app)

    def gen():
        yield b"foo"
        yield b"bar\n"
        yield b"baz"

    response = requests.post(server.url, gen())
    assert response.ok
    assert response.text == "foobar\nbaz"

    assert request_list[0].headers["Transfer-Encoding"].lower() == "chunked"


def test_close_iterable_response(serve_asgi_adapter):
    class IterableResponse:
        def __init__(self, data: list[bytes]):
            self.data = data
            self.closed = False

        def __iter__(self):
            for packet in self.data:
                yield packet

        def close(self):
            # should be called through the werkzeug layers
            self.closed = True

    iterable = IterableResponse([b"foo", b"bar"])

    @Request.application
    def app(request: Request) -> Response:
        return Response(iterable, 200)

    server = serve_asgi_adapter(app)

    response = requests.get(server.url, stream=True)

    gen = response.iter_content(chunk_size=3)
    assert next(gen) == b"foo"
    assert next(gen) == b"bar"
    assert not iterable.closed

    with pytest.raises(StopIteration):
        next(gen)

    assert iterable.closed


def test_input_stream_methods(serve_asgi_adapter):
    @Request.application
    def app(request: Request) -> Response:
        assert request.stream.read(1) == b"f"
        assert request.stream.readline(10) == b"ood\n"
        assert request.stream.readline(3) == b"bar"
        assert next(request.stream) == b"ber\n"
        assert request.stream.readlines(3) == [b"fizz\n"]
        assert request.stream.readline() == b"buzz\n"
        assert request.stream.read() == b"really\ndone"
        assert request.stream.read(10) == b""

        return Response("ok", 200)

    server = serve_asgi_adapter(app)

    def gen():
        yield b"fo"
        yield b"od\n"
        yield b"barber\n"
        yield b"fizz\n"
        yield b"buzz\n"
        yield b"really\n"
        yield b"done"

    response = requests.post(server.url, data=gen())
    assert response.ok
    assert response.text == "ok"


def test_input_stream_readlines(serve_asgi_adapter):
    @Request.application
    def app(request: Request) -> Response:
        assert request.stream.readlines() == [b"fizz\n", b"buzz\n", b"done"]
        return Response("ok", 200)

    server = serve_asgi_adapter(app)

    def gen():
        yield b"fizz\n"
        yield b"buzz\n"
        yield b"done"

    response = requests.post(server.url, data=gen())
    assert response.ok
    assert response.text == "ok"


def test_input_stream_readlines_with_limit(serve_asgi_adapter):
    @Request.application
    def app(request: Request) -> Response:
        assert request.stream.readlines(1000) == [b"fizz\n", b"buzz\n", b"done"]
        return Response("ok", 200)

    server = serve_asgi_adapter(app)

    def gen():
        yield b"fizz\n"
        yield b"buzz\n"
        yield b"done"

    response = requests.post(server.url, data=gen())
    assert response.ok
    assert response.text == "ok"


def test_multipart_post(serve_asgi_adapter):
    @Request.application
    def app(request: Request) -> Response:
        assert request.mimetype == "multipart/form-data"

        result = {}
        for k, file_storage in request.files.items():
            result[k] = file_storage.stream.read().decode("utf-8")

        return Response(json.dumps(result), 200)

    server = serve_asgi_adapter(app)

    response = requests.post(server.url, files={"foo": "bar", "baz": "ed"})
    assert response.ok
    assert response.json() == {"foo": "bar", "baz": "ed"}


def test_multipart_post_large_payload(serve_asgi_adapter):
    @Request.application
    def app(request: Request) -> Response:
        try:
            assert request.mimetype == "multipart/form-data"

            result = {}
            for k, file_storage in request.files.items():
                result[k] = len(file_storage.stream.read())

            return Response(json.dumps(result), 200)
        except Exception:
            LOG.exception("error")
            raise

    server = serve_asgi_adapter(app)

    payload = (
        "\0" * 70_000
    )  # there's a chunk size of 65536 configured in werkzeug which is what we're testing here

    response = requests.post(server.url, files={"file": payload})
    assert response.ok
    assert response.json() == {"file": 70_000}


def test_utf8_path(serve_asgi_adapter):
    @Request.application
    def app(request: Request) -> Response:
        assert request.path == "/foo/Ā0Ä"
        assert request.environ["PATH_INFO"] == "/foo/Ä\x800Ã\x84"

        return Response("ok", 200)

    server = serve_asgi_adapter(app)

    response = requests.get(server.url + "/foo/Ā0Ä")
    assert response.ok


def test_serve_multiple_apps(serve_asgi_adapter):
    @Request.application
    def app0(request: Request) -> Response:
        return Response("ok0", 200)

    @Request.application
    def app1(request: Request) -> Response:
        return Response("ok1", 200)

    server0 = serve_asgi_adapter(app0)
    server1 = serve_asgi_adapter(app1)

    executor = ThreadPoolExecutor(6)

    response0_ftr = executor.submit(requests.get, server0.url)
    response1_ftr = executor.submit(requests.get, server1.url)
    response2_ftr = executor.submit(requests.get, server0.url)
    response3_ftr = executor.submit(requests.get, server1.url)
    response4_ftr = executor.submit(requests.get, server0.url)
    response5_ftr = executor.submit(requests.get, server1.url)

    executor.shutdown()

    result0 = response0_ftr.result(timeout=2)
    assert result0.ok
    assert result0.text == "ok0"
    result1 = response1_ftr.result(timeout=2)
    assert result1.ok
    assert result1.text == "ok1"
    result2 = response2_ftr.result(timeout=2)
    assert result2.ok
    assert result2.text == "ok0"
    result3 = response3_ftr.result(timeout=2)
    assert result3.ok
    assert result3.text == "ok1"
    result4 = response4_ftr.result(timeout=2)
    assert result4.ok
    assert result4.text == "ok0"
    result5 = response5_ftr.result(timeout=2)
    assert result5.ok
    assert result5.text == "ok1"


def test_lifespan_listener(serve_asgi_adapter):
    events = Queue()

    @Request.application
    def app(request: Request) -> Response:
        events.put("request")
        return Response("ok", 200)

    class LifespanListener(ASGILifespanListener):
        def on_startup(self):
            events.put("startup")

        def on_shutdown(self):
            events.put("shutdown")

    listener = LifespanListener()

    server = serve_asgi_adapter(app, listener)

    assert events.get(timeout=5) == "startup"
    assert events.qsize() == 0

    assert requests.get(server.url).ok

    assert events.get(timeout=5) == "request"
    assert events.qsize() == 0

    server.shutdown()

    assert events.get(timeout=5) == "shutdown"
    assert events.qsize() == 0


def test_websocket_listener(serve_asgi_adapter):
    class WebsocketApp:
        def __call__(self, environ: WebSocketEnvironment):
            ws: ASGIWebSocketAdapter = environ["asgi.websocket"]
            event = ws.asgi_receive()
            assert event["type"] == "websocket.connect"
            ws.asgi_send(
                {
                    "type": "websocket.accept",
                    "subprotocol": None,
                    "headers": [(b"X-Foo-Bar", b"foobar")],
                }
            )

            event = ws.asgi_receive()
            assert event == {"type": "websocket.receive", "text": "hello world", "bytes": None}

            ws.asgi_send(
                {
                    "type": "websocket.send",
                    "text": f"echo: {event['text']}",
                }
            )

            ws.asgi_send({"type": "websocket.close", "code": 1000, "reason": "test done"})

    server = serve_asgi_adapter(None, websocket_listener=WebsocketApp())

    client = websocket.WebSocket()
    client.connect(server.url.replace("http://", "ws://"))
    assert client.handshake_response.status == 101
    assert client.handshake_response.headers
    assert client.handshake_response.headers["connection"] == "Upgrade"
    assert client.handshake_response.headers["x-foo-bar"] == "foobar"

    client.send("hello world")
    assert client.recv() == "echo: hello world"
    client.close()



import io

import pytest
from werkzeug.exceptions import NotFound

from rolo import Response
from tests import static


def test_for_resource_html():
    response = Response.for_resource(static, "index.html")
    assert response.content_type == "text/html; charset=utf-8"
    assert response.get_data() == b'<html lang="en">\n<body>hello</body>\n</html>\n'
    assert response.status == "200 OK"


def test_for_resource_txt():
    response = Response.for_resource(static, "test.txt")
    assert response.content_type == "text/plain; charset=utf-8"
    assert response.get_data() == b"hello world\n"
    assert response.status == "200 OK"


def test_for_resource_with_custom_response_status_and_headers():
    response = Response.for_resource(static, "test.txt", status=201, headers={"X-Foo": "Bar"})
    assert response.content_type == "text/plain; charset=utf-8"
    assert response.get_data() == b"hello world\n"
    assert response.status == "201 CREATED"
    assert response.headers.get("X-Foo") == "Bar"


def test_for_resource_not_found():
    with pytest.raises(NotFound):
        Response.for_resource(static, "doesntexist.txt")


def test_for_json():
    response = Response.for_json(
        {"foo": "bar", "420": 69, "isTrue": True},
    )
    assert response.content_type == "application/json"
    assert response.get_data() == b'{"foo": "bar", "420": 69, "isTrue": true}'
    assert response.status == "200 OK"


def test_for_json_with_custom_response_status_and_headers():
    response = Response.for_json(
        {"foo": "bar", "420": 69, "isTrue": True},
        status=201,
        headers={"X-Foo": "Bar"},
    )
    assert response.content_type == "application/json"
    assert response.get_data() == b'{"foo": "bar", "420": 69, "isTrue": true}'
    assert response.status == "201 CREATED"
    assert response.headers.get("X-Foo") == "Bar"


@pytest.mark.parametrize(
    argnames="data",
    argvalues=[
        b"foobar",
        "foobar",
        io.BytesIO(b"foobar"),
        [b"foo", b"bar"],
    ],
)
def test_set_response(data):
    response = Response()
    response.set_response(data)
    assert response.get_data() == b"foobar"


def test_update_from():
    original = Response(
        [b"foo", b"bar"], 202, headers={"X-Foo": "Bar"}, mimetype="application/octet-stream"
    )

    response = Response()
    response.update_from(original)

    assert response.get_data() == b"foobar"
    assert response.status_code == 202
    assert response.headers.get("X-Foo") == "Bar"
    assert response.content_type == "application/octet-stream"



from typing import Any, Dict

import pytest
from werkzeug.exceptions import NotFound

from rolo import Request, Response, Router
from rolo.dispatcher import handler_dispatcher


class TestHandlerDispatcher:
    def test_handler_dispatcher(self):
        router = Router(dispatcher=handler_dispatcher())

        def handler_foo(_request: Request) -> Response:
            return Response("ok")

        def handler_bar(_request: Request, bar, baz) -> Response:
            response = Response()
            response.set_json({"bar": bar, "baz": baz})
            return response

        router.add("/foo", handler_foo)
        router.add("/bar/<int:bar>/<baz>", handler_bar)

        assert router.dispatch(Request("GET", "/foo")).data == b"ok"
        assert router.dispatch(Request("GET", "/bar/420/ed")).json == {"bar": 420, "baz": "ed"}

        with pytest.raises(NotFound):
            assert router.dispatch(Request("GET", "/bar/asfg/ed"))

    def test_handler_dispatcher_invalid_signature(self):
        router = Router(dispatcher=handler_dispatcher())

        def handler(_request: Request, arg1) -> Response:  # invalid signature
            return Response("ok")

        router.add("/foo/<arg1>/<arg2>", handler)

        with pytest.raises(TypeError):
            router.dispatch(Request("GET", "/foo/a/b"))

    def test_handler_dispatcher_with_dict_return(self):
        router = Router(dispatcher=handler_dispatcher())

        def handler(_request: Request, arg1) -> Dict[str, Any]:
            return {"arg1": arg1, "hello": "there"}

        router.add("/foo/<arg1>", handler)
        assert router.dispatch(Request("GET", "/foo/a")).json == {"arg1": "a", "hello": "there"}

    def test_handler_dispatcher_with_list_return(self):
        router = Router(dispatcher=handler_dispatcher())

        def handler(_request: Request, arg1) -> Dict[str, Any]:
            return [{"arg1": arg1, "hello": "there"}, 1, 2, "3", [4, 5]]

        router.add("/foo/<arg1>", handler)
        assert router.dispatch(Request("GET", "/foo/a")).json == [
            {"arg1": "a", "hello": "there"},
            1,
            2,
            "3",
            [4, 5],
        ]

    def test_handler_dispatcher_with_text_return(self):
        router = Router(dispatcher=handler_dispatcher())

        def handler(_request: Request, arg1) -> str:
            return f"hello: {arg1}"

        router.add("/<arg1>", handler)
        assert router.dispatch(Request("GET", "/world")).data == b"hello: world"

    def test_handler_dispatcher_with_none_return(self):
        router = Router(dispatcher=handler_dispatcher())

        def handler(_request: Request):
            return None

        router.add("/", handler)
        assert router.dispatch(Request("GET", "/")).status_code == 200



import typing as t

from werkzeug.datastructures import Headers, MultiDict
from werkzeug.wrappers import Request

from rolo.response import Response

if t.TYPE_CHECKING:
    from _typeshed.wsgi import StartResponse, WSGIEnvironment

import logging

from .gateway import Gateway

LOG = logging.getLogger(__name__)


class WsgiGateway:
    """
    Exposes a ``Gateway`` as a WSGI application.
    """

    gateway: Gateway

    def __init__(self, gateway: Gateway) -> None:
        super().__init__()
        self.gateway = gateway

    def __call__(
        self, environ: "WSGIEnvironment", start_response: "StartResponse"
    ) -> t.Iterable[bytes]:
        # create request from environment
        LOG.debug(
            "%s %s%s",
            environ["REQUEST_METHOD"],
            environ.get("HTTP_HOST"),
            environ.get("RAW_URI"),
        )
        request = Request(environ)

        raw_headers = environ.get("rolo.headers") or environ.get("asgi.headers")
        if raw_headers:
            # restores raw headers from ASGI scope, which allows dashes in header keys
            # see https://github.com/pallets/werkzeug/issues/940

            request.headers = Headers(
                MultiDict([(k.decode("latin-1"), v.decode("latin-1")) for (k, v) in raw_headers])
            )
        else:
            # by default, werkzeug requests from environ are immutable
            request.headers = Headers(request.headers)

        # prepare response
        response = Response()

        self.gateway.process(request, response)

        return response(environ, start_response)



import json
import mimetypes
import typing as t
from importlib import resources

from werkzeug.exceptions import NotFound
from werkzeug.wrappers import Response as WerkzeugResponse

if t.TYPE_CHECKING:
    from types import ModuleType


class Response(WerkzeugResponse):
    """
    An HTTP Response object, which simply extends werkzeug's Response object with a few convenience methods.
    """

    def update_from(self, other: WerkzeugResponse):
        """
        Updates this response object with the data from the given response object. It reads the status code,
        the response data, and updates its own headers (overwrites existing headers, but does not remove ones
        not present in the given object). Also updates ``call_on_close`` callbacks in the same way.

        :param other: the response object to read from
        """
        self.status_code = other.status_code
        self.response = other.response
        self._on_close.extend(other._on_close)
        self.headers.update(other.headers)

    def set_json(self, doc: t.Any, cls: t.Type[json.JSONEncoder] = None):
        """
        Serializes the given dictionary using localstack's ``CustomEncoder`` into a json response, and sets the
        mimetype automatically to ``application/json``.

        :param doc: the response dictionary to be serialized as JSON
        :param cls: the JSON encoder class to use for serializing the passed document
        """
        self.data = json.dumps(doc, cls=cls)
        self.mimetype = "application/json"

    def set_response(self, response: t.Union[str, bytes, bytearray, t.Iterable[bytes]]):
        """
        Function to set the low-level ``response`` object. This is copied from the werkzeug Response constructor. The
        response attribute always holds an iterable of bytes. Passing a str, bytes or bytearray is equivalent to
        calling ``response.data = <response>``. If None is passed, then it will create an empty list. If anything
        else is passed, the value is set directly. This value can be a list of bytes, and iterator that returns bytes
        (e.g., a generator), which can be used by the underlying server to stream responses to the client. Anything else
        (like passing dicts) will result in errors at lower levels of the server.

        :param response: the response value
        """
        if response is None:
            self.response = []
        elif isinstance(response, (str, bytes, bytearray)):
            self.data = response
        else:
            self.response = response

        return self

    def to_readonly_response_dict(self) -> t.Dict:
        """
        Returns a read-only version of a response dictionary as it is often expected by other libraries like boto.
        """
        return {
            "body": self.stream if self.is_streamed else self.data,
            "status_code": self.status_code,
            "headers": dict(self.headers),
        }

    @classmethod
    def for_json(cls, doc: t.Any, *args, **kwargs) -> "Response":
        """
        Creates a new JSON response from the given document. It automatically sets the mimetype to ``application/json``.

        :param doc: the document to serialize into JSON
        :param args: arguments passed to the ``Response`` constructor
        :param kwargs: keyword arguments passed to the ``Response`` constructor
        :return: a new Response object
        """
        response = cls(*args, **kwargs)
        response.set_json(doc)
        return response

    @classmethod
    def for_resource(cls, module: "ModuleType", path: str, *args, **kwargs) -> "Response":
        """
        Looks up the given file in the given module, and creates a new Response object with the contents of that
        file. It guesses the mimetype of the file and sets it in the response accordingly. If the file does not exist
        ,it raises a ``NotFound`` error.

        :param module: the module to look up the file in
        :param path: the path/file name
        :return: a new Response object
        """
        resource = resources.files(module).joinpath(path)
        if not resource.is_file():
            raise NotFound()
        mimetype = mimetypes.guess_type(resource.name)
        mimetype = mimetype[0] if mimetype and mimetype[0] else "application/octet-stream"

        return cls(resource.open("rb"), *args, mimetype=mimetype, **kwargs)



class WebSocketError(IOError):
    """Base class for websocket errors"""

    pass


class WebSocketDisconnectedError(WebSocketError):
    """Raised when the client has disconnected while the server is still trying to receive data."""

    default_code = 1005
    """https://asgi.readthedocs.io/en/latest/specs/www.html#disconnect-receive-event-ws"""

    def __init__(self, code: int = None):
        self.code = code if code is not None else self.default_code
        super().__init__(f"Websocket disconnected code={self.code}")


class WebSocketProtocolError(WebSocketError):
    """Raised if there is a problem in the interaction between app and the websocket server."""

    pass



"""Adapter API between high-level rolo.websocket.request and an underlying IO frameworks like ASGI or
twisted."""
import dataclasses
import typing as t

from werkzeug.datastructures import Headers

WebSocketEnvironment: t.TypeAlias = t.Dict[str, t.Any]
"""Special WSGIEnvironment that has a ``rolo.websocket`` key that stores a `Websocket` instance."""


class Event:
    """A websocket event (subset of ``wsproto.events``)."""

    pass


@dataclasses.dataclass
class Message(Event):
    data: bytes | str


@dataclasses.dataclass
class TextMessage(Message):
    data: str


@dataclasses.dataclass
class BytesMessage(Message):
    data: bytes


@dataclasses.dataclass
class CreateConnection(Event):
    """
    This indicates the first event of the websocket after a connection upgrade. For example, in wsproto
    this corresponds to a ``Request`` event, or ``websocket.connect`` event in ASGI.
    """

    pass


@dataclasses.dataclass
class AcceptConnection(Event):
    subprotocol: t.Optional[str] = None
    extensions: list[str] = dataclasses.field(default_factory=list)
    extra_headers: list[tuple[bytes, bytes]] = dataclasses.field(default_factory=list)


class WebSocketAdapter:
    """
    Adapter to plug the high-level interfaces ``WebSocket`` and ``WebSocketRequest`` into an IO framework.
    It doesn't cover the full websocket protocol API (for instance there are no Ping/Pong events),
    under the assumption that the lower-level IO framework will abstract them away.
    """

    def accept(
        self,
        subprotocol: str = None,
        extensions: list[str] = None,
        extra_headers: Headers = None,
        timeout: float = None,
    ):
        """
        Accept the websocket upgrade request and send an accept message back to the client. This or
        ``reject`` must be the first things to be called.

        :param subprotocol: the accepted subprotocol
        :param extensions: any accepted extensions to use
        :param extra_headers: headers to pass to the accept response
        :param timeout: optional timeout
        """
        raise NotImplementedError

    def reject(
        self,
        status_code: int,
        headers: Headers = None,
        body: t.Iterable[bytes] = None,
        timeout: float = None,
    ):
        """
        Reject the websocket request. This means sending an actual HTTP response back to the client, i.e.,
        not upgrading the connection. This only makes sense before any call to ``receive`` was made.

        :param status_code: the HTTP response status code
        :param headers: the HTTP response headers
        :param body: the body
        :param timeout: optional timeout
        """
        raise NotImplementedError

    def receive(
        self,
        timeout: float = None,
    ) -> CreateConnection | Message:
        """Blocking IO method to wait for the next ``Message`` or, if not initialized yet, the first
        ``CreateConnection`` event."""
        raise NotImplementedError

    def send(
        self,
        event: Message,
        timeout: float = None,
    ):
        """
        Send the given message to the websocket.

        :param event: the message to send
        :param timeout: optional timeout
        :return:
        """
        raise NotImplementedError

    def close(self, code: int = 1001, reason: str = None, timeout: float = None):
        """
        If the underlying websocket connection has already been closed, this call is ignore, so it's safe
        to always call.
        """
        raise NotImplementedError


class WebSocketListener(t.Protocol):
    """
    Similar protocol to a WSGIApplication, only it expects a Websocket instead of a WSGIEnvironment.
    """

    def __call__(self, environ: WebSocketEnvironment):
        """
        Called when a new Websocket connection is established. To initiate the connection, you need to perform
        the connect handshake yourself. First, receive the ``websocket.connect`` event, and then send the
        ``websocket.accept`` event. Here's a minimal example::

            def accept(self, environ: WebsocketEnvironment):
                websocket: WebSocketAdapter = environ['rolo.websocket']
                event = websocket.receive()
                if isinstance(event, CreateConnection):
                    websocket.accept()
                else:
                    websocket.close(1002)  # protocol error
                    return

                while True:
                    event = websocket.receive()
                    if isinstance(event, CloseConnection):
                            return
                    print(event)

        In reality, you wouldn't be using the websocket adapter directly, the server would probably create a
        ``rolo.websocket.WebSocketRequest`` and serve it accordingly through a ``Gateway``.

        :param environ: The new Websocket environment
        """
        raise NotImplementedError



import asyncio
import concurrent.futures.thread
from asyncio import AbstractEventLoop
from typing import Optional

from rolo.asgi import ASGIAdapter, ASGILifespanListener
from rolo.websocket.request import WebSocketRequest

from .gateway import Gateway
from .wsgi import WsgiGateway


class _ThreadPool(concurrent.futures.thread.ThreadPoolExecutor):
    """
    This thread pool executor removes the threads it creates from the global ``_thread_queues`` of
    ``concurrent.futures.thread``, which joins all created threads at python exit and will block interpreter shutdown if
    any threads are still running, even if they are daemon threads.
    """

    def _adjust_thread_count(self) -> None:
        super()._adjust_thread_count()

        for t in self._threads:
            if not t.daemon:
                continue
            try:
                del concurrent.futures.thread._threads_queues[t]
            except KeyError:
                pass


class AsgiGateway:
    """
    Exposes a Gateway as an ASGI3 application. Under the hood, it uses a WsgiGateway with a threading async/sync bridge.
    """

    default_thread_count = 1000

    gateway: Gateway

    def __init__(
        self,
        gateway: Gateway,
        event_loop: Optional[AbstractEventLoop] = None,
        threads: int = None,
        lifespan_listener: Optional[ASGILifespanListener] = None,
        websocket_listener=None,
    ) -> None:
        self.gateway = gateway

        self.event_loop = event_loop or asyncio.get_event_loop()
        self.executor = _ThreadPool(
            threads or self.default_thread_count, thread_name_prefix="asgi_gw"
        )
        self.adapter = ASGIAdapter(
            WsgiGateway(gateway),
            event_loop=event_loop,
            executor=self.executor,
            lifespan_listener=lifespan_listener,
            websocket_listener=websocket_listener or WebSocketRequest.listener(gateway.accept),
        )
        self._closed = False

    async def __call__(self, scope, receive, send) -> None:
        """
        ASGI3 application interface.

        :param scope: the ASGI request scope
        :param receive: the receive callable
        :param send: the send callable
        """
        if self._closed:
            raise RuntimeError("Cannot except new request on closed ASGIGateway")

        return await self.adapter(scope, receive, send)

    def close(self):
        """
        Close the ASGIGateway by shutting down the underlying executor.
        """
        self._closed = True
        self.executor.shutdown(wait=False, cancel_futures=True)



"""rolo pytest plugin used both for internal testing and as testing library."""
import asyncio
import dataclasses
import socket
import threading
import time
import typing
from typing import Protocol

import pytest
from werkzeug import Request as WerkzeugRequest
from werkzeug import serving

from rolo import Router
from rolo.asgi import ASGIAdapter, ASGILifespanListener
from rolo.dispatcher import handler_dispatcher
from rolo.gateway import Gateway
from rolo.gateway.asgi import AsgiGateway
from rolo.gateway.wsgi import WsgiGateway
from rolo.serving.twisted import TwistedGateway
from rolo.websocket.adapter import WebSocketListener

if typing.TYPE_CHECKING:
    from hypercorn.typing import ASGIFramework


class ServerInfo(Protocol):
    url: str
    host: str
    port: int


class Server(ServerInfo):
    def shutdown(self):
        ...


@dataclasses.dataclass
class _ServerInfo:
    host: str
    port: int
    url: str


@pytest.fixture
def serve_wsgi_app():
    servers: list[serving.BaseWSGIServer] = []

    def _serve(app, host: str = "localhost", port: int = None) -> serving.BaseWSGIServer | Server:
        srv = serving.make_server(host, port or 0, app, threaded=True)
        name = threading._newname("test-server-%d")
        threading.Thread(target=srv.serve_forever, name=name, daemon=True).start()
        servers.append(srv)
        srv.url = f"http://{srv.host}:{srv.port}"
        return srv

    yield _serve

    for server in servers:
        server.shutdown()


@pytest.fixture
def wsgi_router_server(serve_wsgi_app) -> tuple[Router, serving.BaseWSGIServer | Server]:
    """Creates a new Router with a handler dispatcher, serves it through a newly created ASGI server, and returns
    both the router and the server.
    """
    router = Router(dispatcher=handler_dispatcher())
    app = WerkzeugRequest.application(router.dispatch)
    return router, serve_wsgi_app(app)


@pytest.fixture()
def serve_asgi_app():
    import hypercorn
    import hypercorn.asyncio

    _server_shutdown = []

    def _create(
        app: "ASGIFramework",
        config: hypercorn.Config = None,
        event_loop: asyncio.AbstractEventLoop = None,
    ) -> Server:
        host = "localhost"
        port = get_random_tcp_port()
        bind = f"localhost:{port}"

        if not config:
            config = hypercorn.Config()
            config.h11_pass_raw_headers = True
            config.bind = [bind]

        event_loop = event_loop or asyncio.new_event_loop()
        close = asyncio.Event()
        closed = threading.Event()

        async def _set_close():
            close.set()

        def _run():
            event_loop.run_until_complete(
                hypercorn.asyncio.serve(app, config, shutdown_trigger=close.wait)
            )
            closed.set()

        def _shutdown():
            if close.is_set():
                return
            asyncio.run_coroutine_threadsafe(_set_close(), event_loop)
            closed.wait(timeout=10)
            try:
                app.close()
            except AttributeError:
                pass
            asyncio.run_coroutine_threadsafe(event_loop.shutdown_asyncgens(), event_loop)
            event_loop.shutdown_default_executor()
            event_loop.stop()
            event_loop.close()

        _server_shutdown.append(_shutdown)
        threading.Thread(
            target=_run, name=threading._newname("asgi-server-%d"), daemon=True
        ).start()

        srv = _ServerInfo(host, port, f"http://{host}:{port}")
        srv.shutdown = _shutdown

        assert wait_server_is_up(srv), f"gave up waiting for server {srv}"

        return srv

    yield _create

    for server_shutdown in _server_shutdown:
        server_shutdown()


@pytest.fixture()
def serve_asgi_adapter(serve_asgi_app):
    def _create(
        wsgi_app,
        lifespan_listener: ASGILifespanListener = None,
        websocket_listener: WebSocketListener = None,
    ):
        loop = asyncio.new_event_loop()
        return serve_asgi_app(
            ASGIAdapter(
                wsgi_app,
                event_loop=loop,
                lifespan_listener=lifespan_listener,
                websocket_listener=websocket_listener,
            ),
            event_loop=loop,
        )

    yield _create


@pytest.fixture
def serve_wsgi_gateway(serve_wsgi_app):
    def _serve(gateway: Gateway) -> Server:
        return serve_wsgi_app(WsgiGateway(gateway))

    return _serve


@pytest.fixture
def serve_asgi_gateway(serve_asgi_app):
    def _serve(gateway: Gateway) -> Server:
        loop = asyncio.new_event_loop()
        return serve_asgi_app(AsgiGateway(gateway, event_loop=loop), event_loop=loop)

    return _serve


@pytest.fixture(scope="session")
def twisted_reactor():
    """Session fixture that controls the lifecycle of the main twisted reactor."""
    from twisted.internet import reactor
    from twisted.internet.error import ReactorAlreadyRunning
    from twisted.web.http import HTTPFactory

    def _run():
        if reactor.running:
            return

        try:
            # for some reason, when using a `SelectReactor` (like you do by default on MacOS), whatever
            # protocols are added to the reactor via `listenTCP` _after_ `run` has been called,
            # are not served properly. We see this because the request calls in tests block forever. If we
            # add any listener here before calling `run`, then for some reason it works. 🤷
            reactor.listenTCP(get_random_tcp_port(), HTTPFactory())
            reactor.run(installSignalHandlers=False)
        except ReactorAlreadyRunning:
            pass

    threading.Thread(target=_run, daemon=True).start()

    assert poll_condition(
        lambda: reactor.running, timeout=5
    ), f"gave up waiting for {reactor} to start"

    yield reactor

    reactor.stop()


@pytest.fixture
def serve_twisted_tcp_server(twisted_reactor):
    """Factory ficture for serving a twisted protocol factory (like ``Site``) through the twisted reactor."""
    from twisted.internet.tcp import Port

    ports: list[Port] = []

    def _create(protocol_factory):
        port = get_random_tcp_port()
        host = "localhost"
        ports.append(twisted_reactor.listenTCP(port, protocol_factory))
        srv = _ServerInfo(host, port, f"http://{host}:{port}")
        assert wait_server_is_up(srv), f"gave up waiting for {srv}"
        return srv

    yield _create

    for _port in ports:
        _port.stopListening()


@pytest.fixture
def serve_twisted_gateway(serve_twisted_tcp_server):
    def _create(gateway):
        return serve_twisted_tcp_server(TwistedGateway(gateway))

    yield _create


@pytest.fixture
def serve_twisted_websocket_listener(twisted_reactor, serve_twisted_tcp_server):
    from twisted.web.server import Site

    from rolo.serving.twisted import HeaderPreservingWSGIResource, WebsocketResourceDecorator

    def _create(websocket_listener: WebSocketListener):
        site = Site(
            WebsocketResourceDecorator(
                original=HeaderPreservingWSGIResource(
                    twisted_reactor, twisted_reactor.getThreadPool(), None
                ),
                websocketListener=websocket_listener,
            )
        )
        return serve_twisted_tcp_server(site)

    return _create


def is_server_up(srv: ServerInfo):
    args = socket.getaddrinfo(srv.host, srv.port, socket.AF_INET, socket.SOCK_STREAM)
    for family, socktype, proto, _canonname, sockaddr in args:
        s = socket.socket(family, socktype, proto)
        try:
            s.connect(sockaddr)
        except socket.error:
            return False
        else:
            s.close()
            return True


def wait_server_is_up(srv: ServerInfo, timeout: float = 10, interval: float = 0.1) -> bool:
    return poll_condition(lambda: is_server_up(srv), timeout=timeout, interval=interval)


def get_random_tcp_port() -> int:
    import socket

    sock = socket.socket()
    sock.bind(("", 0))
    return sock.getsockname()[1]


def poll_condition(
    condition: typing.Callable[[], bool],
    timeout: float = None,
    interval: float = 0.5,
) -> bool:
    """
    Poll evaluates the given condition until a truthy value is returned. It does this every `interval` seconds
    (0.5 by default), until the timeout (in seconds, if any) is reached.

    Poll returns True once `condition()` returns a truthy value, or False if the timeout is reached.
    """
    remaining = 0
    if timeout is not None:
        remaining = timeout

    while not condition():
        if timeout is not None:
            remaining -= interval

            if remaining <= 0:
                return False

        time.sleep(interval)

    return True



import pytest
from werkzeug.exceptions import MethodNotAllowed

from rolo import Request, Resource, Response, Router, resource
from rolo.dispatcher import handler_dispatcher


class TestResource:
    def test_resource_decorator_dispatches_correctly(self):
        router = Router(dispatcher=handler_dispatcher())

        requests = []

        @resource("/_localstack/health")
        class TestResource:
            def on_get(self, req):
                requests.append(req)
                return "GET/OK"

            def on_post(self, req):
                requests.append(req)
                return {"ok": "POST"}

            def on_head(self, req):
                # this is ignored
                requests.append(req)
                return "HEAD/OK"

        router.add(TestResource())

        request1 = Request("GET", "/_localstack/health")
        request2 = Request("POST", "/_localstack/health")
        request3 = Request("HEAD", "/_localstack/health")
        assert router.dispatch(request1).get_data(True) == "GET/OK"
        assert router.dispatch(request1).get_data(True) == "GET/OK"
        assert router.dispatch(request2).json == {"ok": "POST"}
        assert router.dispatch(request3).get_data(True) == "HEAD/OK"
        assert len(requests) == 4
        assert requests[0] is request1
        assert requests[1] is request1
        assert requests[2] is request2
        assert requests[3] is request3

    def test_resource_dispatches_correctly(self):
        router = Router(dispatcher=handler_dispatcher())

        class TestResource:
            def on_get(self, req):
                return "GET/OK"

            def on_post(self, req):
                return "POST/OK"

            def on_head(self, req):
                return "HEAD/OK"

        router.add(Resource("/_localstack/health", TestResource()))

        request1 = Request("GET", "/_localstack/health")
        request2 = Request("POST", "/_localstack/health")
        request3 = Request("HEAD", "/_localstack/health")
        assert router.dispatch(request1).get_data(True) == "GET/OK"
        assert router.dispatch(request2).get_data(True) == "POST/OK"
        assert router.dispatch(request3).get_data(True) == "HEAD/OK"

    def test_dispatch_to_non_existing_method_raises_exception(self):
        router = Router(dispatcher=handler_dispatcher())

        @resource("/_localstack/health")
        class TestResource:
            def on_post(self, request):
                return "POST/OK"

        router.add(TestResource())

        with pytest.raises(MethodNotAllowed):
            assert router.dispatch(Request("GET", "/_localstack/health"))
        assert router.dispatch(Request("POST", "/_localstack/health")).get_data(True) == "POST/OK"

    def test_resource_with_default_dispatcher(self):
        router = Router()

        @resource("/_localstack/<path>")
        class TestResource:
            def on_get(self, req, args):
                return Response.for_json({"message": "GET/OK", "path": args["path"]})

            def on_post(self, req, args):
                return Response.for_json({"message": "POST/OK", "path": args["path"]})

        router.add(TestResource())
        assert router.dispatch(Request("GET", "/_localstack/health")).json == {
            "message": "GET/OK",
            "path": "health",
        }
        assert router.dispatch(Request("POST", "/_localstack/foobar")).json == {
            "message": "POST/OK",
            "path": "foobar",
        }

    def test_resource_overwrite_with_resource_wrapper(self):
        router = Router(dispatcher=handler_dispatcher())

        @resource("/_localstack/health")
        class TestResourceHealth:
            def on_get(self, req):
                return Response.for_json({"message": "GET/OK", "path": req.path})

            def on_post(self, req):
                return Response.for_json({"message": "POST/OK", "path": req.path})

        endpoints = TestResourceHealth()
        router.add(endpoints)
        router.add(Resource("/health", endpoints))

        assert router.dispatch(Request("GET", "/_localstack/health")).json == {
            "message": "GET/OK",
            "path": "/_localstack/health",
        }
        assert router.dispatch(Request("POST", "/_localstack/health")).json == {
            "message": "POST/OK",
            "path": "/_localstack/health",
        }

        assert router.dispatch(Request("GET", "/health")).json == {
            "message": "GET/OK",
            "path": "/health",
        }
        assert router.dispatch(Request("POST", "/health")).json == {
            "message": "POST/OK",
            "path": "/health",
        }



"""This module holds glue code between web servers and rolo. Any ASGI-compliant server can server rolo with
a few patches via ``rolo.serving.asgi``. Moreover, rolo can be served through twisted via
``rolo.serving.twisted``."""



from .request import Request
from .resource import Resource, resource
from .response import Response
from .router import Router, route

__all__ = ["route", "resource", "Resource", "Router", "Response", "Request"]



from io import BytesIO
from typing import IO, TYPE_CHECKING, Dict, Mapping, Optional, Tuple, Union
from urllib.parse import quote, unquote, urlencode, urlparse

if TYPE_CHECKING:
    from _typeshed.wsgi import WSGIEnvironment

from werkzeug.datastructures import Headers, MultiDict
from werkzeug.test import encode_multipart
from werkzeug.wrappers.request import Request as WerkzeugRequest


def dummy_wsgi_environment(
    method: str = "GET",
    path: str = "",
    headers: Optional[Union[Dict, Headers]] = None,
    body: Optional[Union[bytes, str, IO[bytes]]] = None,
    scheme: str = "http",
    root_path: str = "/",
    query_string: Optional[str] = None,
    remote_addr: Optional[str] = None,
    server: Optional[Tuple[str, Optional[int]]] = None,
    raw_uri: Optional[str] = None,
) -> "WSGIEnvironment":
    """
    Creates a dummy WSGIEnvironment that represents a standalone sans-IO HTTP requests.

    See https://wsgi.readthedocs.io/en/latest/definitions.html#standard-environ-keys

    :param method: The HTTP request method (such as GET or POST)
    :param path: The remainder of the request URL's path. This may be an empty string, if the
        request URL targets the application root and does not have a trailing slash.
    :param headers: optional HTTP headers
    :param body: the body of the request
    :param scheme: the scheme (http or https)
    :param root_path: The initial portion of the request URL's path that corresponds to the
        application object.
    :param query_string: The portion of the request URL that follows the “?”, if any. May be
        empty or absent.
    :param remote_addr: The address making the request
    :param server: The server (tuple of server name and port)
    :param raw_uri: The original path that may contain url encoded path elements.
    :return: A WSGIEnvironment dictionary
    """

    # Standard environ keys
    environ = {
        "REQUEST_METHOD": method,
        # prepare the paths for the "WSGI decoding dance" done by werkzeug
        "SCRIPT_NAME": unquote(quote(root_path.rstrip("/")), "latin-1"),
        "PATH_INFO": unquote(quote(path), "latin-1"),
        "SERVER_PROTOCOL": "HTTP/1.1",
        "QUERY_STRING": query_string or "",
    }

    if raw_uri:
        if query_string:
            raw_uri += "?" + query_string
        environ["RAW_URI"] = raw_uri
        environ["REQUEST_URI"] = environ["RAW_URI"]

    if server:
        environ["SERVER_NAME"] = server[0]
        if server[1]:
            environ["SERVER_PORT"] = str(server[1])
        else:
            environ["SERVER_PORT"] = "80"
    else:
        environ["SERVER_NAME"] = "127.0.0.1"
        environ["SERVER_PORT"] = "80"

    if remote_addr:
        environ["REMOTE_ADDR"] = remote_addr

    if headers:
        set_environment_headers(environ, headers)

    if not body:
        body = b""

    if isinstance(body, (str, bytes)):
        data = body.encode("utf-8") if isinstance(body, str) else body

        wsgi_input = BytesIO(data)
        if "CONTENT_LENGTH" not in environ:
            # try to determine content length from body
            environ["CONTENT_LENGTH"] = str(len(data))
    else:
        wsgi_input = body

    # WSGI environ keys
    environ["wsgi.version"] = (1, 0)
    environ["wsgi.url_scheme"] = scheme
    environ["wsgi.input"] = wsgi_input
    environ["wsgi.input_terminated"] = True
    environ["wsgi.errors"] = BytesIO()
    environ["wsgi.multithread"] = True
    environ["wsgi.multiprocess"] = False
    environ["wsgi.run_once"] = False

    return environ


def set_environment_headers(environ: "WSGIEnvironment", headers: Union[Dict, Headers]):
    # Collect all the headers to set
    # (this might be is accessing the environment, this needs to be done before removing the items from the env)
    new_headers = {}
    for k, v in headers.items():
        name = k.upper().replace("-", "_")

        if name not in ("CONTENT_TYPE", "CONTENT_LENGTH"):
            name = f"HTTP_{name}"

        val = v
        if name in new_headers:
            val = new_headers[name] + "," + val

        new_headers[name] = val

    # Clear the HTTP headers in the env
    header_keys = [name for name in environ if name.startswith("HTTP_")]
    for name in header_keys:
        environ.pop(name, None)

    # Set the new headers in the env
    for k, v in new_headers.items():
        environ[k] = v


class Request(WerkzeugRequest):
    """
    An HTTP request object. This is (and should remain) a drop-in replacement for werkzeug's WSGI
    compliant Request objects. It allows simple sans-IO requests outside a web server environment.

    DO NOT add methods that are not also part of werkzeug.wrappers.request.Request object.
    """

    def __init__(
        self,
        method: str = "GET",
        path: str = "",
        headers: Union[Mapping, Headers] = None,
        body: Union[bytes, str] = None,
        scheme: str = "http",
        root_path: str = "/",
        query_string: Union[bytes, str] = b"",
        remote_addr: str = None,
        server: Optional[Tuple[str, Optional[int]]] = None,
        raw_path: str = None,
    ):
        # decode query string if necessary (latin-1 is what werkzeug would expect)
        if isinstance(query_string, bytes):
            query_string = query_string.decode("latin-1")

        # create the WSGIEnvironment dictionary that represents this request
        environ = dummy_wsgi_environment(
            method=method,
            path=path,
            headers=headers,
            body=body,
            scheme=scheme,
            root_path=root_path,
            query_string=query_string,
            remote_addr=remote_addr,
            server=server,
            raw_uri=raw_path,
        )

        super(Request, self).__init__(environ)

        # restore originally passed headers:
        # werkzeug normally provides read-only access to headers set in the WSGIEnvironment through the EnvironHeaders
        # class, here we make them mutable again. moreover, WSGI header encoding conflicts with RFC2616. see this github
        # issue for a discussion: https://github.com/pallets/werkzeug/issues/940
        headers = Headers(headers)
        # these two headers are treated separately in the WSGI environment, so we extract them if necessary
        for h in ["content-length", "content-type"]:
            if h not in headers and h in self.headers:
                headers[h] = self.headers[h]
        self.headers = headers

    @classmethod
    def application(cls, *args):
        # werkzeug's application decorator assumes its Request constructor signature, which our Request doesn't support.
        # using ``application`` from our request therefore creates runtime errors. this makes sure no one runs into
        # these problems. if we want to support it, we need to create compatibility with werkzeug's Request constructor
        raise NotImplementedError


def get_raw_path(request) -> str:
    """
    Returns the raw_path inside the request without the query string. The request can either be a Quart Request
    object (that encodes the raw path in request.scope['raw_path']) or a Werkzeug WSGI request (that encodes the raw
    URI in request.environ['RAW_URI']).

    :param request: the request object
    :return: the raw path if any
    """
    if hasattr(request, "environ"):
        # werkzeug/flask request (already a string, and contains the query part)
        # we need to parse it, because the RAW_URI can contain a full URL if it is specified in the HTTP request
        raw_uri: str = request.environ.get("RAW_URI", "")
        if raw_uri.startswith("//"):
            # if the RAW_URI starts with double slashes, `urlparse` will fail to decode it as path only
            # it also means that we already only have the path, so we just need to remove the query string
            return raw_uri.split("?")[0]
        return urlparse(raw_uri or request.path).path

    if hasattr(request, "scope"):
        # quart request raw_path comes as bytes, and without the query part
        return request.scope.get("raw_path", request.path).decode("utf-8")

    raise ValueError("cannot extract raw path from request object %s" % request)


def get_full_raw_path(request: WerkzeugRequest) -> str:
    """
    Returns the full raw request path (with original URL encoding), including the query string.
    This is _not_ equal to request.url, since there the path section would be url-encoded while the query part will be
    (partly) url-decoded.
    """
    query_str = f"?{request.query_string.decode('latin1')}" if request.query_string else ""
    raw_path = f"{get_raw_path(request)}{query_str}"
    return raw_path


def get_raw_base_url(request: Request) -> str:
    """
    Returns the base URL (with original URL encoding). This does not include the query string.
    This is the encoding-preserving equivalent to `request.base_url`.
    """
    return get_raw_current_url(
        request.scheme, request.host, request.root_path, get_raw_path(request)
    )


def get_raw_current_url(
    scheme: str,
    host: str,
    root_path: Optional[str] = None,
    path: Optional[str] = None,
    query_string: Optional[bytes] = None,
) -> str:
    """
    `werkzeug.sansio.utils.get_current_url` implementation without
    any encoding dances.
    The given paths and query string are directly used without any encodings
    (to avoid any double encodings).
    It can be used to recreate the raw URL for a request.

    :param scheme: The protocol the request used, like ``"https"``.
    :param host: The host the request was made to. See :func:`get_host`.
    :param root_path: Prefix that the application is mounted under. This
        is prepended to ``path``.
    :param path: The path part of the URL after ``root_path``.
    :param query_string: The portion of the URL after the "?".
    """
    url = [scheme, "://", host]

    if root_path is None:
        url.append("/")
        return "".join(url)

    url.append(root_path.rstrip("/"))
    url.append("/")

    if path is None:
        return "".join(url)

    url.append(path.lstrip("/"))

    if query_string:
        url.append("?")
        url.append(query_string)

    return "".join(url)


def restore_payload(request: Request) -> bytes:
    """
    This method takes a request and restores the original payload from it even after it has been consumed. A werkzeug
    request consumes form/multipart data from the stream, and here we are serializing it back to a request a client
    could make . This is a useful method to have when we are proxying requests, i.e., create outgoing requests from
    incoming requests that were subject to previous parsing.

    TODO: this construct is not great and will definitely be a source of trouble. but something like this will
      inevitably become the the basis of proxying werkzeug requests. The alternative is to build our own request object
      that memoizes the original payload before parsing.
    """
    if request.shallow:
        return b""

    data = request.data

    if request.method != "POST":
        return data

    if request.mimetype == "multipart/form-data":
        boundary = request.content_type.split("=")[1]

        fields = MultiDict()
        fields.update(request.form)
        fields.update(request.files)

        _, data_files = encode_multipart(fields, boundary)
        data += data_files

    elif request.mimetype == "application/x-www-form-urlencoded":
        data += urlencode(list(request.form.items(multi=True))).encode("utf-8")

    return data



import json
from typing import Any, Dict, Protocol, Type, Union

from werkzeug import Response as WerkzeugResponse

from .request import Request
from .response import Response
from .router import Dispatcher, RequestArguments

ResultValue = Union[
    WerkzeugResponse,
    str,
    bytes,
    Dict[str, Any],  # a JSON dict
    list[Any],
]


def _populate_response(
    response: WerkzeugResponse, result: ResultValue, json_encoder: Type[json.JSONEncoder]
):
    if result is None:
        return response

    elif isinstance(result, (str, bytes, bytearray)):
        response.data = result
    elif isinstance(result, (dict, list)):
        response.data = json.dumps(result, cls=json_encoder)
        response.mimetype = "application/json"
    else:
        raise ValueError("unhandled result type %s", type(result))

    return response


class Handler(Protocol):
    """
    A protocol used by a ``Router`` together with the dispatcher created with ``handler_dispatcher``. Endpoints added
    with this protocol take as first argument the HTTP request object, and then as keyword arguments the request
    parameters added in the rule. This makes it work very similar to flask routes.

    Example code could look like this::

        def my_route(request: Request, organization: str, repo: str):
            return {"something": "returned as json response"}

        router = Router(dispatcher=handler_dispatcher)
        router.add("/<organization>/<repo>", endpoint=my_route)

    """

    def __call__(self, request: Request, **kwargs) -> ResultValue:
        """
        Handle the given request.

        :param request: the HTTP request object
        :param kwargs: the url request parameters
        :return: a string or bytes value, a dict to create a json response, or a raw werkzeug Response object.
        """
        raise NotImplementedError


def handler_dispatcher(json_encoder: Type[json.JSONEncoder] = None) -> Dispatcher[Handler]:
    """
    Creates a Dispatcher that treats endpoints like callables of the ``Handler`` Protocol.

    :param json_encoder: optionally the json encoder class to use for translating responses
    :return: a new dispatcher
    """

    def _dispatch(request: Request, endpoint: Handler, args: RequestArguments) -> Response:
        result = endpoint(request, **args)
        if isinstance(result, WerkzeugResponse):
            return result
        response = Response()
        if result is not None:
            _populate_response(response, result, json_encoder)
        return response

    return _dispatch



"""
This module enables the resource class pattern, where each respective ``on_<http_method>`` method of a class is
treated like an endpoint for the respective HTTP method. The following shows an example of how the pattern is used::

    class Foo:
        def on_get(self, request: Request):
            return {"ok": "GET it"}

        def on_post(self, request: Request):
            return {"ok": "it was POSTed"}


    router = Router(dispatcher=resource_dispatcher())
    router.add(Resource("/foo", Foo())
"""
from typing import Any, Iterable, Optional, Type

from werkzeug.routing import Map, Rule, RuleFactory

from .router import route

_resource_methods = [
    "on_head",  # it's important that HEAD rules are added first (werkzeug matching order)
    "on_get",
    "on_post",
    "on_put",
    "on_patch",
    "on_delete",
    "on_options",
    "on_trace",
]


def resource(path: str, host: Optional[str] = None, **kwargs):
    """
    Class decorator that turns every method that follows the pattern ``on_<http-method>`` into a route,
    where the allowed method for that route is automatically set to the method specified in the function name. Example
    when using a Router with the ``handler_dispatcher``::

        @resource("/myresource/<resource_id>")
        class MyResource:
            def on_get(request: Request, resource_id: str) -> Response:
                return Response(f"GET called on {resource_id}")

            def on_post(request: Request, resource_id: str) -> Response:
                return Response(f"POST called on {resource_id}")

    This class can then be added to a router via ``router.add_route_endpoints(MyResource())``.

    Note that, if an on_get method is present in the resource but on_head is not, then HEAD requests are automatically
    routed to ``on_get``. This replicates Werkzeug's behavior https://werkzeug.palletsprojects.com/en/2.2.x/routing/.

    :param path: the path pattern to match
    :param host: an optional host matching pattern. if not pattern is given, the rule matches any host
    :param kwargs: any other argument that can be passed to ``werkzeug.routing.Rule``
    :return: a class where each matching function is wrapped as a ``_RouteEndpoint``
    """
    kwargs.pop("methods", None)

    def _wrapper(cls: Type):
        for name in _resource_methods:
            member = getattr(cls, name, None)
            if member is None:
                continue

            http_method = name[3:].upper()
            setattr(cls, name, route(path, host, methods=[http_method], **kwargs)(member))

        return cls

    return _wrapper


class Resource(RuleFactory):
    """
    Exposes a given object that follows the "Resource" class pattern as a ``RuleFactory` that can then be added to a
    Router. Example use when using a Router with the ``handler_dispatcher``::

        class MyResource:
            def on_get(request: Request, resource_id: str) -> Response:
                return Response(f"GET called on {resource_id}")

            def on_post(request: Request, resource_id: str) -> Response:
                return Response(f"POST called on {resource_id}")

        router.add(Resource("/myresource/<resource_id>", MyResource()))

    Note that, if an on_get method is present in the resource but on_head is not, then HEAD requests are automatically
    routed to ``on_get``. This replicates Werkzeug's behavior https://werkzeug.palletsprojects.com/en/2.2.x/routing/.
    """

    def __init__(self, path: str, obj: Any, host: Optional[str] = None, **kwargs):
        self.path = path
        self.obj = obj
        self.host = host
        self.kwargs = kwargs

    def get_rules(self, map: "Map") -> Iterable["Rule"]:
        rules = []
        for name in _resource_methods:
            member = getattr(self.obj, name, None)
            if member is None:
                continue

            http_method = name[3:].upper()
            rules.append(
                Rule(
                    self.path, endpoint=member, methods=[http_method], host=self.host, **self.kwargs
                )
            )
        return rules



import threading
from typing import List, Tuple

import pytest
import requests
import werkzeug
from werkzeug.exceptions import MethodNotAllowed, NotFound
from werkzeug.routing import RequestRedirect, Submount

from rolo import Request, Response, Router
from rolo.router import E, RequestArguments, RuleAdapter, WithHost, route


def noop(*args, **kwargs):
    """Test dispatcher that does nothing"""
    return Response()


def echo_params_json(request: Request, params: dict[str, str]):
    """Test dispatcher that echoes the url match parameters as json"""
    r = Response()
    r.set_json(params)
    return r


class RequestCollector:
    """Test dispatcher that collects requests into a list"""

    requests: List[Tuple[Request, E, RequestArguments]]

    def __init__(self) -> None:
        super().__init__()
        self.requests = []

    def __call__(self, request: Request, endpoint: E, args: RequestArguments) -> Response:
        self.requests.append((request, endpoint, args))
        return Response()


class TestRouter:
    # these are sanity check for the router and dispatching logic. since the matching is done by werkzeug's Map,
    # there is no need for thorough testing URL matching.

    def test_dispatch_raises_not_found(self):
        router = Router()
        router.add("/foobar", noop)
        with pytest.raises(NotFound):
            assert router.dispatch(Request("GET", "/foo"))

    def test_default_dispatcher_invokes_correct_endpoint(self):
        router = Router()

        def index(_: Request, args) -> Response:
            response = Response()
            response.set_json(args)
            return response

        def users(_: Request, args) -> Response:
            response = Response()
            response.set_json(args)
            return response

        router.add("/", index)
        router.add("/users/<int:user_id>", users)

        assert router.dispatch(Request("GET", "/")).json == {}
        assert router.dispatch(Request("GET", "/users/12")).json == {"user_id": 12}

    def test_dispatch_with_host_matching(self):
        router = Router()

        def ep_all(_: Request, args) -> Response:
            response = Response()
            response.set_json(dict(method="all", **args))
            return response

        def ep_index1(_: Request, args) -> Response:
            response = Response()
            response.set_json(dict(method="1", **args))
            return response

        def ep_index2(_: Request, args) -> Response:
            response = Response()
            response.set_json(dict(method="2", **args))
            return response

        router.add("/", ep_index1, host="localhost:<port>")
        router.add("/", ep_index2, host="localhost:12345")
        router.add("/all", ep_all, host="<host>")

        def invoke(path, server, port):
            return router.dispatch(Request("GET", path, server=(server, port))).json

        assert invoke("/", "localhost", 4566) == {"method": "1", "port": "4566"}
        assert invoke("/", "localhost", 12345) == {"method": "2"}
        assert invoke("/all", "127.0.0.1", None) == {"method": "all", "host": "127.0.0.1"}
        assert invoke("/all", "127.0.0.1", 12345) == {"method": "all", "host": "127.0.0.1:12345"}

        with pytest.raises(NotFound):
            invoke("/", "localstack.cloud", None)

    def test_custom_dispatcher(self):
        collector = RequestCollector()
        router = Router(dispatcher=collector)

        router.add("/", "index")
        router.add("/users/<int:id>", "users")

        router.dispatch(Request("GET", "/"))
        router.dispatch(Request("GET", "/users/12"))

        _, endpoint, args = collector.requests[0]
        assert endpoint == "index"
        assert args == {}

        _, endpoint, args = collector.requests[1]
        assert endpoint == "users"
        assert args == {"id": 12}

    def test_regex_path_dispatcher(self):
        router = Router()
        rgx = r"([^.]+)endpoint(.*)"
        regex = f"/<regex('{rgx}'):dist>/"
        router.add(path=regex, endpoint=noop)
        assert router.dispatch(Request(method="GET", path="/test-endpoint"))
        with pytest.raises(NotFound):
            router.dispatch(Request(method="GET", path="/test-not-point"))

    def test_regex_host_dispatcher(self):
        router = Router()
        rgx = r"\.cloudfront.(net|localhost\.localstack\.cloud)"
        router.add(path="/", endpoint=noop, host=f"<dist_id><regex('{rgx}'):host>:<port>")
        assert router.dispatch(
            Request(
                method="GET",
                headers={"Host": "ad91f538.cloudfront.localhost.localstack.cloud:5446"},
            )
        )
        with pytest.raises(NotFound):
            router.dispatch(
                Request(
                    method="GET",
                    headers={"Host": "ad91f538.cloudfront.amazon.aws.com:5446"},
                )
            )

    def test_port_host_dispatcher(self):
        collector = RequestCollector()
        router = Router(dispatcher=collector)
        router.add(path="/", endpoint=noop, host="localhost.localstack.cloud<port:port>")
        # matches with the port!
        assert router.dispatch(
            Request(
                method="GET",
                headers={"Host": "localhost.localstack.cloud:4566"},
            )
        )
        assert collector.requests.pop()[2] == {"port": 4566}
        # matches without the port!
        assert router.dispatch(
            Request(
                method="GET",
                headers={"Host": "localhost.localstack.cloud"},
            )
        )
        assert collector.requests.pop()[2] == {"port": None}

        # invalid port
        with pytest.raises(NotFound):
            router.dispatch(
                Request(
                    method="GET",
                    headers={"Host": "localhost.localstack.cloud:544a6"},
                )
            )

        # does not match the host
        with pytest.raises(NotFound):
            router.dispatch(
                Request(
                    method="GET",
                    headers={"Host": "localstack.cloud:5446"},
                )
            )

    def test_path_converter(self):
        router = Router()
        router.add(path="/<path:path>", endpoint=echo_params_json)

        assert router.dispatch(Request(path="/my")).json == {"path": "my"}
        assert router.dispatch(Request(path="/my/")).json == {"path": "my/"}
        assert router.dispatch(Request(path="/my//path")).json == {"path": "my//path"}
        assert router.dispatch(Request(path="/my//path/")).json == {"path": "my//path/"}
        assert router.dispatch(Request(path="/my/path foobar")).json == {"path": "my/path foobar"}
        assert router.dispatch(Request(path="//foobar")).json == {"path": "foobar"}
        assert router.dispatch(Request(path="//foobar/")).json == {"path": "foobar/"}

    def test_path_converter_with_args(self):
        router = Router()
        router.add(path="/with-args/<some_id>/<path:path>", endpoint=echo_params_json)

        assert router.dispatch(Request(path="/with-args/123456/my")).json == {
            "some_id": "123456",
            "path": "my",
        }

        # werkzeug no longer removes trailing slashes in matches
        assert router.dispatch(Request(path="/with-args/123456/my/")).json == {
            "some_id": "123456",
            "path": "my/",
        }

        # works with sub paths
        assert router.dispatch(Request(path="/with-args/123456/my/path")).json == {
            "some_id": "123456",
            "path": "my/path",
        }

        # no sub path raises 404
        with pytest.raises(NotFound):
            router.dispatch(Request(path="/with-args/123456"))

        with pytest.raises(NotFound):
            router.dispatch(Request(path="/with-args/123456/"))

        # with the default slash behavior of the URL map (merge_slashes=False), werkzeug tries to redirect
        # the call to /with-args/123456/my/ (note: this is desirable for web servers, not always for us
        # though)
        with pytest.raises(RequestRedirect):
            assert router.dispatch(Request(path="/with-args/123456//my/"))

    def test_path_converter_and_regex_converter_in_host(self):
        router = Router()
        router.add(
            path="/<path:path>",
            host="foobar.us-east-1.opensearch.localhost.localstack.cloud<regex('(?::.*)?'):port>",
            endpoint=echo_params_json,
        )
        assert router.dispatch(
            Request(
                method="GET",
                path="/_cluster/health",
                headers={"Host": "foobar.us-east-1.opensearch.localhost.localstack.cloud:4566"},
            )
        ).json == {"path": "_cluster/health", "port": ":4566"}

    def test_path_converter_and_port_converter_in_host(self):
        router = Router()
        router.add(
            path="/<path:path>",
            host="foobar.us-east-1.opensearch.localhost.localstack.cloud<port:port>",
            endpoint=echo_params_json,
        )
        assert router.dispatch(
            Request(
                method="GET",
                path="/_cluster/health",
                headers={"Host": "foobar.us-east-1.opensearch.localhost.localstack.cloud:4566"},
            )
        ).json == {"path": "_cluster/health", "port": 4566}

        assert router.dispatch(
            Request(
                method="GET",
                path="/_cluster/health",
                headers={"Host": "foobar.us-east-1.opensearch.localhost.localstack.cloud"},
            )
        ).json == {"path": "_cluster/health", "port": None}

    def test_path_converter_and_greedy_regex_in_host(self):
        router = Router()
        router.add(
            path="/<path:path>",
            # note how the regex '.*' will also include the port (so port will not do anything)
            host="foobar.us-east-1.opensearch.<regex('.*'):host><port:port>",
            endpoint=echo_params_json,
        )
        assert router.dispatch(
            Request(
                method="GET",
                path="/_cluster/health",
                headers={"Host": "foobar.us-east-1.opensearch.localhost.localstack.cloud:4566"},
            )
        ).json == {
            "path": "_cluster/health",
            "host": "localhost.localstack.cloud:4566",
            "port": None,
        }

    def test_remove_rule(self):
        router = Router()

        def index(_: Request, args) -> Response:
            return Response(b"index")

        def users(_: Request, args) -> Response:
            return Response(b"users")

        rule0 = router.add("/", index)
        rule1 = router.add("/users/<int:user_id>", users)

        assert router.dispatch(Request("GET", "/")).data == b"index"
        assert router.dispatch(Request("GET", "/users/12")).data == b"users"

        router.remove(rule1)

        assert router.dispatch(Request("GET", "/")).data == b"index"
        with pytest.raises(NotFound):
            assert router.dispatch(Request("GET", "/users/12"))

        router.remove(rule0)
        with pytest.raises(NotFound):
            assert router.dispatch(Request("GET", "/"))
        with pytest.raises(NotFound):
            assert router.dispatch(Request("GET", "/users/12"))

    def test_remove_rules(self):
        router = Router()

        class MyRoutes:
            @route("/a")
            @route("/a2")
            def route_a(self, request, args):
                return Response(b"a")

            @route("/b")
            def route_b(self, request, args):
                return Response(b"b")

        rules = router.add(MyRoutes())

        assert router.dispatch(Request("GET", "/a")).data == b"a"
        assert router.dispatch(Request("GET", "/a2")).data == b"a"
        assert router.dispatch(Request("GET", "/b")).data == b"b"

        router.remove(rules)

        with pytest.raises(NotFound):
            assert router.dispatch(Request("GET", "/a"))

        with pytest.raises(NotFound):
            assert router.dispatch(Request("GET", "/a2"))

        with pytest.raises(NotFound):
            assert router.dispatch(Request("GET", "/b"))

    def test_remove_non_existing_rule(self):
        router = Router()

        def index(_: Request, args) -> Response:
            return Response(b"index")

        rule = router.add("/", index)
        router.remove(rule)

        with pytest.raises(KeyError) as e:
            router.remove(rule)
        e.match("no such rule")

    def test_router_route_decorator(self):
        router = Router()

        @router.route("/users")
        @router.route("/alternative-users")
        def user(_: Request, args):
            assert not args
            return Response("user")

        @router.route("/users/<int:user_id>")
        def user_id(_: Request, args):
            assert args
            return Response(f"{args['user_id']}")

        assert router.dispatch(Request("GET", "/users")).data == b"user"
        assert router.dispatch(Request("GET", "/alternative-users")).data == b"user"
        assert router.dispatch(Request("GET", "/users/123")).data == b"123"

    def test_add_route_endpoint_with_object(self):
        class MySuperApi:
            @route("/users")
            def user(self, _: Request, args):
                # should be inherited
                assert not args
                return Response("user")

        class MyApi(MySuperApi):
            @route("/users/<int:user_id>")
            def user_id(self, _: Request, args):
                assert args
                return Response(f"{args['user_id']}")

            def foo(self, _: Request, args):
                # should be ignored
                raise NotImplementedError

        api = MyApi()
        router = Router()
        rules = router.add(api)
        assert len(rules) == 2

        assert router.dispatch(Request("GET", "/users")).data == b"user"
        assert router.dispatch(Request("GET", "/users/123")).data == b"123"

    def test_add_route_endpoint_with_object_per_method(self):
        # tests whether there can be multiple rules with different methods to the same URL
        class MyApi:
            @route("/my_api", methods=["GET"])
            def do_get(self, request: Request, _args):
                # should be inherited
                return Response(f"{request.path}/do-get")

            @route("/my_api", methods=["HEAD"])
            def do_head(self, request: Request, _args):
                # should be inherited
                return Response(f"{request.path}/do-head")

            @route("/my_api", methods=["POST", "PUT"])
            def do_post(self, request: Request, _args):
                # should be inherited
                return Response(f"{request.path}/do-post-or-put")

        api = MyApi()
        router = Router()
        rules = router.add(api)
        assert len(rules) == 3

        assert router.dispatch(Request("GET", "/my_api")).data == b"/my_api/do-get"
        assert router.dispatch(Request("HEAD", "/my_api")).data == b"/my_api/do-head"
        assert router.dispatch(Request("POST", "/my_api")).data == b"/my_api/do-post-or-put"
        assert router.dispatch(Request("PUT", "/my_api")).data == b"/my_api/do-post-or-put"

        with pytest.raises(MethodNotAllowed):
            router.dispatch(Request("DELETE", "/my_api"))

    def test_head_requests_are_routed_to_get_handlers(self):
        @route("/my_api", methods=["GET"])
        def do_get(request: Request, _args):
            # should be inherited
            return Response(f"{request.path}/do-get")

        router = Router()
        router.add(do_get)

        assert router.dispatch(Request("GET", "/my_api")).data == b"/my_api/do-get"
        assert router.dispatch(Request("HEAD", "/my_api")).data == b"/my_api/do-get"

    def test_submount_rule_adapter(self):
        @route("/my_api", methods=["GET"])
        def do_get(request: Request, _args):
            # should be inherited
            return Response(f"{request.path}/do-get")

        def hello(request: Request, _args):
            return Response("hello world")

        router = Router()

        # base endpoints
        endpoints = RuleAdapter([do_get, RuleAdapter("/hello", hello)])

        router.add([endpoints, Submount("/foo", [endpoints])])

        assert router.dispatch(Request("GET", "/foo/my_api")).data == b"/foo/my_api/do-get"
        assert router.dispatch(Request("GET", "/my_api")).data == b"/my_api/do-get"

        assert router.dispatch(Request("GET", "/foo/hello")).data == b"hello world"
        assert router.dispatch(Request("GET", "/hello")).data == b"hello world"

    def test_with_host_and_submount(self):
        @route("/my_api", methods=["GET"])
        def do_get(request: Request, _args):
            response = Response()
            response.set_json({"path": request.path, "host": request.host})
            return response

        router = Router()

        router.add(
            [
                WithHost(
                    "foo.localhost.localstack.cloud:4566",
                    [RuleAdapter(do_get)],
                ),
                Submount(
                    "/foo",
                    [RuleAdapter(do_get)],
                ),
            ]
        )

        request = Request("GET", "/foo/my_api")
        assert router.dispatch(request).json == {
            "host": "127.0.0.1",
            "path": "/foo/my_api",
        }

        request = Request("GET", "/my_api", server=("foo.localhost.localstack.cloud", 4566))
        assert router.dispatch(request).json == {
            "path": "/my_api",
            "host": "foo.localhost.localstack.cloud:4566",
        }

        request = Request("GET", "/my_api", server=("localhost.localstack.cloud", 4566))
        with pytest.raises(NotFound):
            router.dispatch(request)


class TestWsgiIntegration:
    def test_with_werkzeug(self):
        # setup up router
        router = Router()

        def index(_: Request, args) -> Response:
            return Response(b"index")

        def echo_json(request: Request, args) -> Response:
            response = Response()
            response.set_json(request.json)
            return response

        def users(_: Request, args) -> Response:
            response = Response()
            response.set_json(args)
            return response

        router.add("/", index)
        router.add("/users/<int:user_id>", users, host="<host>:<port>")
        router.add("/echo/", echo_json, methods=["POST"])

        # serve router through werkzeug
        @werkzeug.Request.application
        def app(request: werkzeug.Request) -> werkzeug.Response:
            return router.dispatch(request)

        host = "localhost"
        server = werkzeug.serving.make_server(host, 0, app=app, threaded=True)
        port = server.port
        url = f"http://{host}:{port}"
        t = threading.Thread(target=server.serve_forever)
        t.start()

        try:
            resp = requests.get(f"{url}/")
            assert resp.ok
            assert resp.content == b"index"

            resp = requests.get(f"{url}/users/123")
            assert resp.ok
            assert resp.json() == {"user_id": 123, "host": host, "port": str(port)}

            resp = requests.get(f"{url}/users")
            assert not resp.ok

            resp = requests.post(f"{url}/echo", json={"foo": "bar", "a": 420})
            assert resp.ok
            assert resp.json() == {"foo": "bar", "a": 420}
        finally:
            server.shutdown()
            t.join(timeout=10)



"""This module contains code to make ASGI play nice with WSGI."""
import asyncio
import io
import logging
import math
import typing as t
from asyncio import AbstractEventLoop
from concurrent.futures import Executor
from io import BufferedReader, RawIOBase
from urllib.parse import quote, unquote, urlparse

from werkzeug.datastructures import Headers

from rolo.websocket import (
    WebSocketDisconnectedError,
    WebSocketEnvironment,
    WebSocketListener,
    WebSocketProtocolError,
)
from rolo.websocket import adapter as rolows

if t.TYPE_CHECKING:
    from _typeshed import WSGIApplication, WSGIEnvironment
    from hypercorn.typing import (
        ASGIReceiveCallable,
        ASGISendCallable,
        HTTPScope,
        Scope,
        WebsocketAcceptEvent,
        WebsocketCloseEvent,
        WebsocketConnectEvent,
        WebsocketDisconnectEvent,
        WebsocketReceiveEvent,
        WebsocketResponseBodyEvent,
        WebsocketResponseStartEvent,
        WebsocketScope,
        WebsocketSendEvent,
    )

    _WebsocketResponse = t.Union[
        WebsocketAcceptEvent,
        WebsocketSendEvent,
        WebsocketResponseStartEvent,
        WebsocketResponseBodyEvent,
        WebsocketCloseEvent,
    ]

    _WebsocketRequest = t.Union[
        WebsocketConnectEvent,
        WebsocketReceiveEvent,
        WebsocketDisconnectEvent,
    ]

LOG = logging.getLogger(__name__)


def populate_wsgi_environment(
    environ: t.Union["WSGIEnvironment", WebSocketEnvironment],
    scope: t.Union["HTTPScope", "WebsocketScope"],
):
    """
    Adds the non-IO parts (e.g., excluding wsgi.input) from the ASGI HTTPScope to the WSGI Environment. See
    WSGI Compatibility for more information on why this works:
    https://asgi.readthedocs.io/en/latest/specs/www.html#wsgi-compatibility

    :param environ: the WSGI environment to populate
    :param scope: the ASGI scope as source
    """
    environ["REQUEST_METHOD"] = scope.get("method", "GET")
    # path/uri info
    # prepare the paths for the "WSGI decoding dance" done by werkzeug
    environ["SCRIPT_NAME"] = unquote(quote(scope.get("root_path", "").rstrip("/")), "latin-1")

    path = scope["path"]
    path = path if path[0] == "/" else urlparse(path).path
    environ["PATH_INFO"] = unquote(quote(path), "latin-1")

    query_string = scope.get("query_string")
    if query_string:
        raw_uri = scope["raw_path"] + b"?" + query_string
        environ["QUERY_STRING"] = query_string.decode("latin1")
    else:
        raw_uri = scope["raw_path"]
        environ["QUERY_STRING"] = ""

    environ["RAW_URI"] = environ["REQUEST_URI"] = raw_uri.decode("utf-8")

    # server address / host
    server = scope.get("server") or ("localhost", 80)
    environ["SERVER_NAME"] = server[0]
    environ["SERVER_PORT"] = str(server[1]) if server[1] else "80"

    # http version
    environ["SERVER_PROTOCOL"] = "HTTP/" + scope["http_version"]

    # client (remote) address
    client = scope.get("client")
    if client:
        environ["REMOTE_ADDR"] = client[0]
        environ["REMOTE_PORT"] = str(client[1])

    # headers
    for name, value in scope["headers"]:
        key = name.decode("latin1").upper().replace("-", "_")

        if key not in ["CONTENT_TYPE", "CONTENT_LENGTH"]:
            key = f"HTTP_{key}"

        environ[key] = value.decode("latin1")

    # wsgi specific keys
    environ["wsgi.version"] = (1, 0)
    environ["wsgi.url_scheme"] = scope.get("scheme", "http")
    environ["wsgi.errors"] = io.BytesIO()
    environ["wsgi.multithread"] = True
    environ["wsgi.multiprocess"] = False
    environ["wsgi.run_once"] = False

    # rolo.headers: a custom key to allow downstream applications to circumvent WSGI header processing. these headers
    # should preserve the original casing as the client sends them.
    headers = scope.get("headers")
    environ["rolo.headers"] = headers
    environ["asgi.headers"] = environ["rolo.headers"]


class _AsyncGeneratorWrapper:
    def __init__(
        self,
        it: t.Iterator,
        loop: t.Optional[AbstractEventLoop] = None,
        executor: t.Optional[Executor] = None,
    ):
        """
        Wraps a given synchronous Iterator as an async generator, where each invocation to ``next(it)``
        will be wrapped in a coroutine execution.

        :param it: the iterator to wrap
        :param loop: the event loop to run the next invocations
        :param executor: the executor to run the synchronous code
        """
        self.it = it
        self.loop = loop or asyncio.get_event_loop()
        self.executor = executor

    def _next_sync(self):
        try:
            return next(self.it)
        except StopIteration:
            pass
        raise StopAsyncIteration

    def __aiter__(self):
        return self

    async def __anext__(self):
        val = await self.loop.run_in_executor(self.executor, self._next_sync)
        return val

    async def aclose(self):
        if close := getattr(self.it, "close", None):
            return await self.loop.run_in_executor(self.executor, close)


def create_wsgi_input(
    receive: "ASGIReceiveCallable", event_loop: t.Optional[AbstractEventLoop] = None
) -> t.IO[bytes]:
    """
    Factory for exposing an ASGIReceiveCallable as an IO stream.

    :param receive: the receive callable
    :param event_loop: the event loop used by the event stream adapter
    :return: a new IO stream that wraps the given receive callable.
    """
    return BufferedReader(RawHTTPRequestEventStreamAdapter(receive, event_loop))


class RawHTTPRequestEventStreamAdapter(RawIOBase):
    """
    An adapter to expose an ASGIReceiveCallable coroutine that returns HTTPRequestEvent instances as an IO
    stream for synchronous WSGI/Werkzeug code. The adapter is a Raw IO stream, meaning it does not have
    optimized ``read``, ``readline``, or ``readlines`` methods. Make sure to use a ``BufferedReader`` around
    the stream adapter.
    """

    def __init__(
        self, receive: "ASGIReceiveCallable", event_loop: t.Optional[AbstractEventLoop] = None
    ) -> None:
        super().__init__()
        self.receive = receive
        self.event_loop = event_loop or asyncio.get_event_loop()

        # internal state
        self._more_body = True
        self._buffered_body = None
        self._buffered_body_pos = 0

    def readable(self) -> bool:
        return True

    def readinto(self, buf: bytearray | memoryview) -> int:
        if not self._more_body:
            return 0

        # max bytes we can write into the buffer
        buf_size = len(buf)

        # _buffered_body holds the carry-over of what we didn't read in the last iteration
        if self._buffered_body is None:
            # read from the underlying socket stream
            recv_future = asyncio.run_coroutine_threadsafe(self.receive(), self.event_loop)
            event = recv_future.result()
            # TODO: disconnect events
            more = event.get("more_body", False)

            if not more:
                self._more_body = False
                return 0

            body = self._buffered_body = event["body"]
            pos = self._buffered_body_pos = 0
        else:
            body = self._buffered_body
            pos = self._buffered_body_pos

        remaining = len(body) - pos

        if remaining <= buf_size:
            # the easiest case, where we write the entire remaining event body into the buffer. we may return
            # less than the buffer size allows, but that's ok for raw IO streams.
            buf[:remaining] = body[pos:]
            self._buffered_body = None
            return remaining

        # in this case, we can read at max buf_size from the body into the buffer, and need to save the
        # rest for the next call
        buf[:buf_size] = body[pos : pos + buf_size]
        self._buffered_body_pos = pos + buf_size

        return buf_size


class WsgiStartResponse:
    """
    A wrapper that exposes an async ``ASGISendCallable`` as synchronous a WSGI ``StartResponse`` protocol callable.
    See this stackoverflow post for a good explanation: https://stackoverflow.com/a/16775731/804840.
    """

    def __init__(
        self,
        send: "ASGISendCallable",
        event_loop: AbstractEventLoop = None,
    ):
        self.send = send
        self.event_loop = event_loop or asyncio.get_event_loop()
        self.sent = 0
        self.content_length = math.inf
        self.finalized = False
        self.started = False

    def __call__(
        self, status: str, headers: t.List[t.Tuple[str, str]], exec_info=None
    ) -> t.Callable[[bytes], t.Any]:
        return self.start_response_sync(status, headers, exec_info)

    def start_response_sync(
        self, status: str, headers: t.List[t.Tuple[str, str]], exec_info=None
    ) -> t.Callable[[bytes], t.Any]:
        """
        The WSGI start_response protocol.

        :param status: the HTTP status (e.g., ``200 OK``) to write
        :param headers: the HTTP headers to write
        :param exec_info: ignored
        :return: a callable that lets you write bytes to the response body
        """
        send = self.send
        loop = self.event_loop

        # start sending response
        asyncio.run_coroutine_threadsafe(
            send(
                {
                    "type": "http.response.start",
                    "status": int(status[:3]),
                    "headers": [(h[0].encode("latin1"), h[1].encode("latin1")) for h in headers],
                }
            ),
            loop,
        ).result()

        self.started = True
        # find out content length if set
        self.content_length = math.inf  # unknown content-length
        for k, v in headers:
            if k.lower() == "content-length":
                self.content_length = int(v)
                break

        return self.write_sync

    def write_sync(self, data: bytes) -> None:
        return asyncio.run_coroutine_threadsafe(self.write(data), self.event_loop).result()

    async def write(self, data: bytes) -> None:
        if not self.started:
            raise ValueError("not started the response yet")
        if getattr(self.send.__self__, "closed", None):
            # the connection has been closed from the client side, set finalized=True to avoid sending more responses
            self.finalized = True
            raise BrokenPipeError("Connection closed")
        await self.send({"type": "http.response.body", "body": data, "more_body": True})
        self.sent += len(data)
        if self.sent >= self.content_length:
            await self.close()

    async def close(self):
        if not self.started:
            raise ValueError("not started the response yet")

        if not self.finalized:
            self.finalized = True
            await self.send({"type": "http.response.body", "body": b"", "more_body": False})


class ASGILifespanListener:
    """
    Simple event handler that is attached to the ASGIAdapter and called on ASGI lifespan events. See
    https://asgi.readthedocs.io/en/latest/specs/lifespan.html.
    """

    def on_startup(self):
        pass

    def on_shutdown(self):
        pass


class ASGIWebSocketAdapter(rolows.WebSocketAdapter):
    """
    Adapter code to serve a ``rolo.websocket.WebSocketRequest`` through ASGI.
    """

    _scope: "WebsocketScope"
    _receive: "ASGIReceiveCallable"
    _send: "ASGISendCallable"

    def __init__(
        self,
        scope: "WebsocketScope",
        receive: "ASGIReceiveCallable",
        send: "ASGISendCallable",
        loop: AbstractEventLoop,
    ):
        self._scope = scope
        self._receive = receive
        self._send = send
        self._loop = loop

    async def asgi_send_async(self, event: "_WebsocketResponse"):
        await self._send(event)

    async def asgi_receive_async(self) -> "_WebsocketRequest":
        return await self._receive()

    def asgi_send(self, event: "_WebsocketResponse", timeout: float = None) -> None:
        """
        Sends an event to the Websocket. Events can be:

        - websocket.accept: https://asgi.readthedocs.io/en/latest/specs/www.html#accept-send-event
        - websocket.send: https://asgi.readthedocs.io/en/latest/specs/www.html#send-send-event
        - websocket.close: https://asgi.readthedocs.io/en/latest/specs/www.html#close-send-event

        :param event: The event to send
        :param timeout: The number of seconds to wait for the result of the async call
        """
        return asyncio.run_coroutine_threadsafe(self.asgi_send_async(event), self._loop).result(
            timeout=timeout
        )

    def asgi_receive(self, timeout: float = None) -> "_WebsocketRequest":
        """
        Listens on the websocket and returns the next event. Events can be:

        - websocket.connect: https://asgi.readthedocs.io/en/latest/specs/www.html#connect-receive-event
        - websocket.receive: https://asgi.readthedocs.io/en/latest/specs/www.html#receive-receive-event
        - websocket.disconnect: https://asgi.readthedocs.io/en/latest/specs/www.html#disconnect-receive-event-ws

        :param timeout: The number of seconds to wait for the event
        :return: The received event
        """
        return asyncio.run_coroutine_threadsafe(self.asgi_receive_async(), self._loop).result(
            timeout
        )

    def receive(self, timeout: float = None) -> rolows.CreateConnection | rolows.Message:
        event = self.asgi_receive(timeout)

        # user-facing events
        if event["type"] == "websocket.connect":
            return rolows.CreateConnection()

        if event["type"] == "websocket.receive":
            event: "WebsocketReceiveEvent"
            text = event.get("text")
            if text is not None:
                return rolows.TextMessage(text)

            buf: bytes = event.get("bytes")
            if buf is not None:
                return rolows.BytesMessage(buf)

            raise WebSocketProtocolError(
                "Both bytes and text are None in the websocket.receive event."
            )

        # internal events
        if event["type"] == "websocket.disconnect":
            event: "WebsocketDisconnectEvent"
            raise WebSocketDisconnectedError(event["code"])

    def send(self, event: rolows.Message, timeout: float = None):
        if isinstance(event, rolows.TextMessage):
            asgi_event = {
                "type": "websocket.send",
                "text": event.data,
                "bytes": None,
            }
        elif isinstance(event, rolows.BytesMessage):
            asgi_event = {
                "type": "websocket.send",
                "text": None,
                "bytes": event.data,
            }
        else:
            raise TypeError(f"Unknown message type {event.__class__.__name__}")

        self.asgi_send(asgi_event, timeout=timeout)

    def reject(
        self,
        status_code: int,
        headers: Headers = None,
        body: t.Iterable[bytes] = None,
        timeout: float = None,
    ):
        self.asgi_send(
            {
                "type": "websocket.http.response.start",
                "status": status_code,
                "headers": self._to_asgi_headers(headers) if headers else [],
            },
            timeout=timeout,
        )
        if body:
            for chunk in body:
                self.asgi_send(
                    {
                        "type": "websocket.http.response.body",
                        "body": chunk,
                        "more_body": True,
                    },
                    timeout=timeout,
                )
        self.asgi_send(
            {
                "type": "websocket.http.response.body",
                "body": b"",
                "more_body": False,
            },
            timeout=timeout,
        )

    def accept(
        self,
        subprotocol: str = None,
        extensions: list[str] = None,
        extra_headers: Headers = None,
        timeout: float = None,
    ):
        # TODO: how to deal with extensions on this level?
        self.asgi_send(
            {
                "type": "websocket.accept",
                "subprotocol": subprotocol,
                "headers": self._to_asgi_headers(extra_headers) if extra_headers else [],
            },
            timeout=timeout,
        )

    def close(self, code: int = 1001, reason: str = None, timeout: float = None):
        self.asgi_send(
            {
                "type": "websocket.close",
                "code": code,
                "reason": reason,
            },
            timeout=timeout,
        )

    def _to_asgi_headers(self, headers: Headers) -> t.List[t.Tuple[bytes, bytes]]:
        return [(h[0].encode("latin1"), h[1].encode("latin1")) for h in headers.to_wsgi_list()]


class ASGIAdapter:
    """
    Adapter to expose a WSGIApplication as an ASGI3Application. This allows you to serve synchronous WSGI applications
    through ASGI servers (e.g., Hypercorn).

    IMPORTANT: The ASGIAdapter needs to use the same event loop as the underlying server. If you pass a new event
    loop to the server, you need to also pass it to the ASGIAdapter.

    https://asgi.readthedocs.io/en/latest/specs/main.html
    """

    def __init__(
        self,
        wsgi_app: "WSGIApplication",
        event_loop: AbstractEventLoop = None,
        executor: Executor = None,
        lifespan_listener: ASGILifespanListener = None,
        websocket_listener: WebSocketListener = None,
    ):
        self.wsgi_app = wsgi_app
        self.event_loop = event_loop or asyncio.get_event_loop()
        self.executor = executor
        self.lifespan_listener = lifespan_listener or ASGILifespanListener()
        self.websocket_listener = websocket_listener

    async def __call__(
        self, scope: "Scope", receive: "ASGIReceiveCallable", send: "ASGISendCallable"
    ):
        """
        The ASGI 3 interface. Can only handle HTTP calls.

        :param scope: the connection scope
        :param receive: the receive callable
        :param send: the send callable
        """
        if scope["type"] == "http":
            return await self.handle_http(scope, receive, send)

        if scope["type"] == "lifespan":
            return await self.handle_lifespan(scope, receive, send)

        if scope["type"] == "websocket":
            return await self.handle_websocket(scope, receive, send)

        raise NotImplementedError("Unhandled protocol %s" % scope["type"])

    def to_wsgi_environment(
        self,
        scope: "HTTPScope",
        receive: "ASGIReceiveCallable",
    ) -> "WSGIEnvironment":
        """
        Creates an IO-ready WSGIEnvironment from the given ASGI HTTP call.

        :param scope: the ASGI HTTP Scope
        :param receive: the ASGI callable to receive the HTTP request
        :return: a WSGIEnvironment
        """
        environ: "WSGIEnvironment" = {}
        populate_wsgi_environment(environ, scope)
        # add IO wrappers
        environ["wsgi.input"] = create_wsgi_input(receive, event_loop=self.event_loop)
        # indicate that the stream is EOF terminated per request
        environ["wsgi.input_terminated"] = True
        return environ

    async def handle_http(
        self, scope: "HTTPScope", receive: "ASGIReceiveCallable", send: "ASGISendCallable"
    ):
        env = self.to_wsgi_environment(scope, receive)

        try:
            response = WsgiStartResponse(send, self.event_loop)

            iterable = await self.event_loop.run_in_executor(
                self.executor, self.wsgi_app, env, response
            )
        except Exception as e:
            LOG.error(
                "Error while trying to schedule execution: %s with environment %s",
                e,
                env,
                exc_info=LOG.isEnabledFor(logging.DEBUG),
            )
            raise

        try:
            if iterable:
                # Generators are also Iterators
                if isinstance(iterable, t.Iterator):
                    iterable = _AsyncGeneratorWrapper(iterable)

                if isinstance(iterable, (t.AsyncIterator, t.AsyncIterable)):
                    async for packet in iterable:
                        await response.write(packet)
                else:
                    for packet in iterable:
                        await response.write(packet)
        except ConnectionError as e:
            client_info = "unknown"
            if client := scope.get("client"):
                address, port = client
                client_info = f"{address}:{port}"
            LOG.debug("Error while writing responses: %s (client_info: %s)", e, client_info)
        finally:
            if iterable and hasattr(iterable, "aclose"):
                await iterable.aclose()
            await response.close()

    def to_websocket_environment(
        self,
        scope: "WebsocketScope",
        receive: "ASGIReceiveCallable",
        send: "ASGISendCallable",
    ) -> WebSocketEnvironment:
        """
        Creates an IO-ready pseudo-WSGI environment from the given ASGI Websocket scope.

        :param scope: the websocket scope
        :param receive: receive callable
        :param send: send callable
        :return: a new websocket environment
        """
        environ: WebSocketEnvironment = {}
        populate_wsgi_environment(environ, scope)
        environ["REQUEST_METHOD"] = "WEBSOCKET"
        environ["rolo.websocket"] = ASGIWebSocketAdapter(scope, receive, send, self.event_loop)
        environ["asgi.websocket"] = environ["rolo.websocket"]
        return environ

    async def handle_websocket(
        self, scope: "WebsocketScope", receive: "ASGIReceiveCallable", send: "ASGISendCallable"
    ):
        if not self.websocket_listener:
            raise NotImplementedError("No websocket listener attached")

        # populate a pseudo-WSGI environment with "WEBSOCKET" as method
        # this can later be used to construct a sans-IO Werkzeug request
        environ = self.to_websocket_environment(scope, receive, send)

        try:
            await self.event_loop.run_in_executor(self.executor, self.websocket_listener, environ)
        except Exception as e:
            LOG.error(
                "Error while trying to schedule execution: %s with environment %s",
                e,
                environ,
                exc_info=LOG.isEnabledFor(logging.DEBUG),
            )
            raise

    async def handle_lifespan(
        self, scope: "HTTPScope", receive: "ASGIReceiveCallable", send: "ASGISendCallable"
    ):
        while True:
            message = await receive()
            if message["type"] == "lifespan.startup":
                try:
                    await self.event_loop.run_in_executor(
                        self.executor, self.lifespan_listener.on_startup
                    )
                    await send({"type": "lifespan.startup.complete"})
                except Exception as e:
                    await send({"type": "lifespan.startup.failed", "message": f"{e}"})

            elif message["type"] == "lifespan.shutdown":
                try:
                    await self.event_loop.run_in_executor(
                        self.executor, self.lifespan_listener.on_shutdown
                    )
                    await send({"type": "lifespan.shutdown.complete"})
                except Exception as e:
                    await send({"type": "lifespan.shutdown.failed", "message": f"{e}"})
                return
            else:
                return
